/* MySQL DB to save results and runtimes of parametric algorithms  */
/*==============================================================*/
/* DBMS name:      MySQL 4.0.14                                 */
/* Created on:     8/27/2006                                    */
/*==============================================================*/

/*==============================================================*/
/* Table: ALGORUN (each entry is one single run).               */
/* For deterministic algos, leave seed empty.                   */
/*==============================================================*/
create table FH_ALGORUN
(
    ALGORUN_CONFIG_ID       int unsigned not null,
	SEED					int not null,
	MEASURED_RUNTIME		float,
	MEASURED_RUNLENGTH		int,
	BEST_SOL				float,
	BEST_LENGTH				int,
	SOLVED					enum('SAT','TIMEOUT','UNSAT','CRASHED'),
	DATE					DATETIME,
	OUTPUT_ID				int,
    unique index algorunconfig_id (ALGORUN_CONFIG_ID,SEED)
);


/*==============================================================*/
/* Table: ALGORUN (each entry is one single run).               */
/* For nonparametric algos leave algorithm_config_id empty.     */
/*==============================================================*/
create table FH_ALGORUN_CONFIG
(
    ID                      int unsigned not null AUTO_INCREMENT,
	ALGORITHM				char(100) not null,
	INSTANCE_ID				int unsigned not null,
	ALGORITHM_CONFIG_ID		int unsigned,
	DESIRED_QUAL            DECIMAL(30,15),
	MACHINE_TYPE_NAME		char(50),
	CUTOFF_TIME				DECIMAL(30,15),
	CUTOFF_LENGTH			int,
	COMMAND					TEXT,
	primary key (ID),
	unique index algorun_config(ALGORITHM,INSTANCE_ID,ALGORITHM_CONFIG_ID,MACHINE_TYPE_NAME,CUTOFF_TIME,CUTOFF_LENGTH)
);

/*==============================================================*/
/* Table: SLS4MPE_CONFIGURATION (complete configuration of SLS4MPE)
/*==============================================================*/
create table SLS4MPE_CONFIGURATION
(
    ID                  int unsigned not null AUTO_INCREMENT,
	A         			int not null,
	B				    int not null,
	C					int not null,
	GLS_SMOOTH          DECIMAL(8,6) not null,
	GLS_INTERVAL        int unsigned not null,
	GLS_PENMULT         int unsigned not null,
    GLS_REAL            int not null,
    GLS_ASPIRATION      int not null,
    N                   int not null,
    CF                  DECIMAL(8,2) not null,
    PERT                int not null,
    PFIX                int not null,
    VNS                 int not null,
    P                   int not null,
    PREL                int not null,
    ACC                 int not null,
    AN                  DECIMAL(8,6) not null,
    T                   DECIMAL(8,6) not null,
    primary key (ID),
	unique index configuration(A,B,GLS_SMOOTH,GLS_INTERVAL,GLS_PENMULT,GLS_REAL,GLS_ASPIRATION,CF,PERT,PFIX,VNS,P,PREL,ACC,AN,T) 
);
    #NOT in index: C for caching and N for (G+StS) noise)

/*==============================================================*/
/* Table: GLS4MPE_CONFIGURATION (complete configuration of SLS4MPE)
/*==============================================================*/
create table GLS4MPE_CONFIGURATION
(
    ID                  int unsigned not null AUTO_INCREMENT,
	B				    int not null,
	GLS_SMOOTH          DECIMAL(8,6) not null,
	GLS_INTERVAL        int unsigned not null,
	GLS_PENMULT         int unsigned not null,
    PREPROC_BOUND       int not null,
    primary key (ID),
	unique index configuration(GLS_SMOOTH,GLS_INTERVAL,GLS_PENMULT,PREPROC_BOUND,B) 
);


/*==============================================================*/
/* Table: SPEAR_CONFIGURATION (complete configuration of SPEAR)
/* Cannot put everything into the key :-(
/* 	unique key configuration(VAR_DEC_HEUR,CLAUSE_SORT_HEUR,CLAUSE_DEL_HEUR,PHASE_DEC_HEUR,VARIABLE_DECAY,CLAUSE_DECAY,RESTART_INC,LEARNED_SIZE_FACTOR,LEARNED_CLAUSES_INC,CLAUSE_ACTIVITY_INC,VAR_ACTIVITY_INC,RAND_PHASE_DEC_FREQ,RAND_VAR_DEC_FREQ,RAND_VAR_DEC_SCALING,RAND_PHASE_SCALING,FIRST_RESTART,BIN_RES_TRESH,ENABLE_BIN_RESOLUTION,UPDATE_DEC_QUEUE,USE_PURE_LITERAL_RULE,PURGE_FALSE_LITS), 
/*==============================================================*/
create table SPEAR_CONFIGURATION_1_0
(
    ID                    int unsigned not null AUTO_INCREMENT,
	VAR_DEC_HEUR          int unsigned not null,
	CLAUSE_SORT_HEUR      int unsigned not null,
	RES_ORDER_HEUR	int unsigned,
	CLAUSE_DEL_HEUR		  int unsigned not null,
	PHASE_DEC_HEUR        int unsigned not null,
	RESOLUTION	int unsigned not null,
	VARIABLE_DECAY        DECIMAL(6,2) not null,
	CLAUSE_DECAY          DECIMAL(6,2) not null,
	RESTART_INC           DECIMAL(6,2) not null,
    LEARNED_SIZE_FACTOR   DECIMAL(6,2) not null,
    LEARNED_CLAUSES_INC   DECIMAL(6,2) not null,
    CLAUSE_ACTIVITY_INC   DECIMAL(6,2) not null,
    VAR_ACTIVITY_INC      DECIMAL(6,2) not null,
    RAND_PHASE_DEC_FREQ   DECIMAL(6,2),
    RAND_VAR_DEC_FREQ     DECIMAL(6,2) not null,
    RAND_VAR_DEC_SCALING  DECIMAL(6,2),
    RAND_PHASE_SCALING    DECIMAL(6,2),
    MAX_RES_LIT_INC	DECIMAL(6,2),
    FIRST_RESTART         int not null,
    RES_CUTOFF_CLS	int,
    RES_CUTOFF_LITS	int,
    MAX_RES_RUNS	int,
    UPDATE_DEC_QUEUE      int not null,
    USE_PURE_LITERAL_RULE int not null,
    primary key (ID),
	index configuration(VAR_DEC_HEUR,CLAUSE_SORT_HEUR,CLAUSE_DEL_HEUR,PHASE_DEC_HEUR,VARIABLE_DECAY,CLAUSE_DECAY,RESTART_INC,LEARNED_SIZE_FACTOR,LEARNED_CLAUSES_INC,CLAUSE_ACTIVITY_INC,VAR_ACTIVITY_INC,RAND_PHASE_DEC_FREQ,RAND_VAR_DEC_FREQ,RAND_VAR_DEC_SCALING,RAND_PHASE_SCALING,FIRST_RESTART) 
);


/*==============================================================*/
/* Table: SAPS_CONFIGURATION (complete configuration of SAPS)
/*==============================================================*/
create table SAPS_CONFIGURATION
(
	ID         			int unsigned not null AUTO_INCREMENT,
	ALPHA				DECIMAL(7,4) not null,
	RHO					DECIMAL(6,4) not null,
	SMOOTH_PROB			DECIMAL(6,4) not null,
	RANDOM_WALK_PROB	DECIMAL(6,4) not null,
	primary key (ID),
	unique index configuration (ALPHA,RHO,SMOOTH_PROB,RANDOM_WALK_PROB)
);


/*==============================================================*/
/* Table: SAT4J_CONFIGURATION
/*==============================================================*/
create table SAT4J_CONFIGURATION
(
	ID         			int unsigned not null AUTO_INCREMENT,
	DSF				    enum('org.sat4j.minisat.constraints.ClausalDataStructureCB', 'org.sat4j.minisat.constraints.ClausalDataStructureCBWL', 'org.sat4j.minisat.constraints.ClausalDataStructureWL', 'org.sat4j.minisat.constraints.MixedDataStructureDaniel', 'org.sat4j.minisat.constraints.MixedDataStructureWithBinary', 'org.sat4j.minisat.constraints.MixedDataStructureWithBinaryAndTernary') not null,
	LEARNING			enum('org.sat4j.minisat.learning.ActiveLearning', 'org.sat4j.minisat.learning.FixedLengthLearning', 'org.sat4j.minisat.learning.LimitedLearning', 'org.sat4j.minisat.learning.MiniSATLearning', 'org.sat4j.minisat.learning.NoLearningButHeuristics', 'org.sat4j.minisat.learning.NoLearningNoHeuristics') not null,
	ACTIVITYPERCENT		DECIMAL(4,2),
	MAXLENGTH			int unsigned,
	LIMITLEARN			int unsigned,
	VARORDER			enum('org.sat4j.minisat.orders.JWOrder', 'org.sat4j.minisat.orders.MyOrder', 'org.sat4j.minisat.orders.PureOrder', 'org.sat4j.minisat.orders.VarOrderHeap', 'org.sat4j.minisat.orders.VarOrderHeapObjective') not null,
	PERIOD				int unsigned,
	SIMP				enum('NO_SIMPLIFICATION', 'SIMPLE_SIMPLIFICATION', 'EXPENSIVE_SIMPLIFICATION') not null,
	VARDECAY			DECIMAL(4,2) not null,
	CLADECAY			DECIMAL(6,4) not null,
	CONFLICTBOUNDINCFACTOR		DECIMAL(4,2) not null,
	INITCONFLICTBOUND		int unsigned not null,
	primary key (ID),
	unique index configuration (DSF,LEARNING,ACTIVITYPERCENT,LIMITLEARN,VARORDER,PERIOD,SIMP,VARDECAY,CLADECAY,CONFLICTBOUNDINCFACTOR,INITCONFLICTBOUND)
);
