require "param_reader.rb"
require "global_helper.rb" # for instance hash

$state_to_int = {}
$stripped_state_to_int = {}
$int_to_full_state = []
$int_to_full_state_string = []
$stripped_int_to_stripped_state = []
$full_int_to_stripped_int = []

# =========================================
# This procedure is called at the first encounter of a state and defines its associated integer for the full state and the stripped version.
# It can also be called again later on.
# Returns the int of the full state, to be used everywhere in connection with the caching Hashes/arrays set up here.
# =========================================
def full_state_to_full_int(full_state)
	full_state_as_string = state_string(full_state)
	unless $state_to_int.key?(full_state_as_string)
		state_int = $state_to_int.length
		$state_to_int[full_state_as_string] = state_int
		$int_to_full_state[state_int] = full_state
		$int_to_full_state_string[state_int] = full_state_as_string

		stripped_state = strip_state(full_state)
		stripped_state_as_string = state_string(stripped_state)
		#=== If we have not seen the stripped state before, set it up. (full -> stripped is N:1)
		unless $stripped_state_to_int.key?(stripped_state_as_string)
			stripped_state_int = $stripped_state_to_int.length
			#$stripped_int_to_one_of_its_full_ints[stripped_state_int] = state_int
			$stripped_state_to_int[stripped_state_as_string] = stripped_state_int
		end
		stripped_state_int = $stripped_state_to_int[stripped_state_as_string]
		$full_int_to_stripped_int[state_int] = stripped_state_int

		$stripped_int_to_stripped_state[stripped_state_int] = stripped_state
	end
	return $state_to_int[full_state_as_string]
end

# =========================================
#                  put string to output and outfile
# =========================================
def output(s)
	if @output_level >= 50
		puts s
		@out.puts s
		@out.flush
	end
end

# =========================================
# Get quality of a parameter setting without extra runs.
# =========================================
def eval(full_state_int)
	raise "eval() takes an int argument, not a #{full_state_int.class}" unless full_state_int.is_a?(Integer)
	level = detail(full_state_int)
	if level > 0
		n = @numRunsPerLevelOfDetail[level-1]
		c = @censoringThresholdPerLevelOfDetail[level-1]
		
		stripped_state_int = $full_int_to_stripped_int[full_state_int]
		res = @cachedResultScalars[stripped_state_int][level-1]
	else
		n = 0
		c = 0
		res = 100000000
	end
	return [res, n, c]
end


# =========================================
# Output evaluation of a state
# =========================================
def str(full_state_int, print_state=true)
	val, n, c = eval(full_state_int)
	if print_state
		return "#{$int_to_full_state_string[full_state_int]} (#{val} [based on #{n} runs with cutoff #{c}])" if val 
		return "#{$int_to_full_state_string[full_state_int]} (-- [no runs])"
	else
		return "(#{val} [based on #{n} runs with cutoff #{c}])" if val 
		return "(-- [no runs])"
	end
end


# =========================================
# Output more detailed information for a state.
# =========================================
def output_result(full_state_int)
	stripped_state_int = $full_int_to_stripped_int[full_state_int]

	entriesForObjective = []
	instance_names = []
	for entry in @instances
		if entry["resultForState"].key?(stripped_state_int)
			entriesForObjective << entry 
			instance_names << entry["name"] unless instance_names.include?(entry["name"])
		else
			break
		end
	end
#	raise "FATAL! Number of results #{numEvaluated(state)} does not equal results from the beginning #{entriesForObjective.length}" if ((@approach=~ /focused/) and not numEvaluated(state)==entriesForObjective.length)
	
	level = detail(full_state_int)
	censoringTime = @censoringThresholdPerLevelOfDetail[level-1]
	
	objectives = getAllObjectives(@algo, @run_obj, entriesForObjective, instance_names, stripped_state_int, censoringTime, @cutoff_length)
#	objectives = getObjectivesForComputedInstances(@algo, @run_obj, entriesForObjective, @instances_sorted, state_as_string)
	#instanceHash = makeStandardInstanceHash(@instances, state_string(stripped_state))

	outputObjectives(instance_names, objectives)
end


def isPruned(bound)
	if bound.class.to_s == "Array"
		return true if bound[0] == "pruned"
		raise "bound is array, but doesn't have pruned as first element."
	end
	return false
end

# =========================================
# Adaptive neighbourhood relation that takes into account conditional parameters.
# =========================================
def neighbourhood(full_state_int)
	stripped_state_int = $full_int_to_stripped_int[full_state_int]
	stripped_state = $stripped_int_to_stripped_state[stripped_state_int]
	full_state = $int_to_full_state[full_state_int]
	result = []
	for param in stripped_state.keys
		next if @fixed_ass.key?(param)
		for value in @domain[param]
			unless full_state[param] == value
				new_state = full_state.dup
				new_state[param] = value

				next if forbidden(new_state, @forbidden_combos)
				#=== Cannot convert state Hash to int yet - only do that for states we actually visit because of memory constraints.
				result << [new_state, param, new_state[param], value]
			end
		end
	end
	if result == []
		puts "fixed_ass:"
		p @fixed_ass
		puts "state.keys"
		p state.keys
		raise "neighbourhood can't be empty!!"
	end
	return result
end


# =========================================================================================
# Bundle actions to be taken when a state is visited in the search.
# =========================================================================================
def visit(full_state_int)
	#=== Update statistics. 
	unless @allVisitedStates.key?(full_state_int)
		@allVisitedStates[full_state_int] = {"number" => @allVisitedStates.size, "numVisited"=>0, "totalTime"=>0, "timesLM"=>0, "timesMadeIncumbent"=>0, "iteration"=>@iteration, "str"=>$int_to_full_state[full_state_int]}
		@orderedVisitedStates << full_state_int
		@statesVisited << full_state_int
	end
	@allVisitedStates[full_state_int]["numVisited"] += 1
end


# =========================================================================================
# Output incumbent state is found.
# =========================================================================================
def output_incumbent_state()
	val, n, c = @incumbent_res
	full_state = $int_to_full_state[@incumbent_state_int]

	numeval = $totalEvaluationCount

#	@param_ils_traj_file.print "#{$total_cputime}, #{val}, #{n}, #{@iteration}, #{@numFlip}, "
	@param_ils_traj_file.print "#{$total_cputime}, #{val}, #{n}, #{@iteration}, #{c}, "
	@param_ils_traj_file.puts @params.map{|x|full_state[x]}.join(", ") # output the current values
	@param_ils_traj_file.flush

	output "New Incumbent: #{$total_cputime}, #{val} [#{n}, #{c}]. With state " + @params.map{|x| "#{x}=#{full_state[x]}"}.join(", ") 
	@out.flush
end


# =========================================================================================
# Set a new incumbent state.
# =========================================================================================
def setNewIncumbent(full_state_int, res=nil, n=nil, c=nil)
	raise "setNewIncumbent() takes an int argument, not a #{full_state_int.class}" unless full_state_int.is_a?(Integer)
	@incumbent_state_int = full_state_int
	if res
		@incumbent_res = [res, n, c]
	else
		@incumbent_res = eval(full_state_int)
	end
	output_incumbent_state
	
	unless @allVisitedStates.key?(full_state_int) 
		@allVisitedStates[full_state_int] = {"number" => @allVisitedStates.size, "numVisited"=>0, "totalTime"=>0, "timesLM"=>0, "timesMadeIncumbent"=>0, "iteration"=>@iteration, "str"=>$int_to_full_state[full_state_int]}
		@orderedVisitedStates << full_state_int
	end
	@allVisitedStates[full_state_int]["timesMadeIncumbent"] += 1
end


# =========================================================================================
# Update the precision of the incumbent.
# =========================================================================================
def update_incumbent(new_inc_res=nil)
	new_inc_res = eval(@incumbent_state_int) unless new_inc_res
	if new_inc_res[1] >= @incumbent_res[1] and new_inc_res[2] >= @incumbent_res[2]
		output " Same incumbent, new precision:"
		setNewIncumbent(@incumbent_state_int, new_inc_res[0], new_inc_res[1], new_inc_res[2])
	end
end

# =========================================================================================
# Get the results for instance-seed combinations for a given param config.
# =========================================================================================
def getRunsOnEntriesWithParams(entries, full_state_int, cutoffTimeForState, test=false)
	stripped_state_int = $full_int_to_stripped_int[full_state_int]
	
	#=== Remove precision to match the one of the DB.
	cutoffTimeForState = (cutoffTimeForState*(10**5)).floor * (10**(-5.0))

#	raise "state_string error: strip_state(strip_state(s)) != strip_state(s): \n#{state_string(strip_state(param_state))} != \n#{state_string(param_state)}" unless state_string(param_state) == state_string(strip_state(param_state))
	oldTotalCpuTime = $total_cputime
	
	getAlgoResultsForEntryParamsPairs(@algo, entries.map{|entry| [entry, stripped_state_int]}, cutoffTimeForState, @cutoff_length, @oncluster, @db)

	@allVisitedStates[full_state_int]["totalTime"] += ($total_cputime - oldTotalCpuTime)

	#=== Assert that we now know all the entries. 
	for entry in entries
		unless entry["resultForState"][stripped_state_int][cutoffTimeForState]
			puts "Empty result! " 
			puts "State"
			p stripped_state_int
			puts "Entry"
			p entry
			p entry["resultForState"]
			p entry["resultForState"][stripped_state_int]
			raise "Empty result"
		end
	end

	#=== Compute the objective for each entry.
	results = entries.map{|entry| singleRunObjective(@algo, @run_obj, entry["resultForState"][stripped_state_int][cutoffTimeForState], entry["qual"], entry["reference"], cutoffTimeForState, @cutoff_length)}

	unless test
		if $total_cputime > @lastOutputCPUTime + 10
			@lastOutputCPUTime  = $total_cputime
			output("#{$totalEvaluationCount}/#{@maxEvals}, #{$total_cputime}/#{@tunerTimeout}")
		end
		#=== Keep track of results for each <param, value> pair.
	end
	return results
end


#=========================================================================================
# Add new runs for param_state, for the original FocusedILS.
# Invariants: Before and after this, @incumbent_state_int is one of the states with the most evaluations, and amongst those, it is the best.
# The number of runs is limited by @N.
# =========================================================================================
def addNewRuns(param_state, numNewRuns)
#	numNewRuns.times{|i| incTimeSpentInState(param_state)}
	numNewRuns.times{|i| break if haveToStop(); boundedIncTimeSpentInState(param_state, nil)}
end

# =========================================================================================
# Return the current level of detail used to evaluate the  state.
# =========================================================================================
def detail(full_state_int)
	stripped_state_int = $full_int_to_stripped_int[full_state_int]
	raise "detail() takes an int argument, not a #{full_state_int.class}" unless full_state_int.is_a?(Integer)
#	state_as_string = state_string(state)
	unless @cachedResultScalars.key?(stripped_state_int)
		@cachedResultScalars[stripped_state_int] = []
	end
	return @cachedResultScalars[stripped_state_int].length
end


# =========================================================================================
# Helper function for better. 
# Returns true iff combination_of_results_1 - combination_of_results_2 <= eps, 
# i.e. iff the first state is better or equal to the second one using the lesser level of detail.
# =========================================================================================
def isBetterWithLesserDetail(state1_int, state2_int, equalIsBetter)
#	puts "comparing \n#{state1} and \n#{state2}"
	level1 = detail(state1_int)
	level2 = detail(state2_int)
	
	minLevel = [level1, level2].min
	return isBetterWithLevel(state1_int, state2_int, equalIsBetter, minLevel)
end

# =========================================================================================
# Returns true iff combination_of_results_1 - combination_of_results_2 <= eps, 
# i.e. iff the first state is better or equal to the second one using that level of detail.
# =========================================================================================
def isBetterWithLevel(state1_int, state2_int, equalIsBetter, level)
	raise "Level can't be <= 0: #{level}" if level <= 0
#	p level

	res1 = @cachedResultScalars[$full_int_to_stripped_int[state1_int]][level-1]
	res2 = @cachedResultScalars[$full_int_to_stripped_int[state2_int]][level-1]
	
#	p res1
#	p res2
#	p equalIsBetter
	
#	p @cachedResultScalars[state_as_string1]
#	p @cachedResultScalars[state_as_string2]
	
	#=== In case res1 was pruned before let's compute it now (enough for the comparison).
	if isPruned(res1)
		if isPruned(res2)
			doRunsBounded(state2_int, level-1, nil)
			res2 = @cachedResultScalars[$full_int_to_stripped_int[state2_int]][level-1]
		end
		doRunsBounded(state1_int, level-1, res2)
		res1 = @cachedResultScalars[$full_int_to_stripped_int[state1_int]][level-1]
	end

	#=== In case res2 was pruned before let's compute it now (enough for the comparison).
	if isPruned(res2)
		doRunsBounded(state2_int, level-1, res1)
		res2 = @cachedResultScalars[$full_int_to_stripped_int[state2_int]][level-1]
	end

	return shortBetter(res1, res2, equalIsBetter)
end


def shortBetter(res1, res2, equalIsBetter=true)
	return false if isPruned(res1) and not isPruned(res2)
	return true if isPruned(res2) and not isPruned(res1)
	
	#=== Both results are pruned, break the tie by the number of solved instances with X times the bound of the incumbent.
	if isPruned(res1) and isPruned(res2)
		return false if res1[1] < res2[1]
		return true if res1[1] > res2[1]
		return "tie" if equalIsBetter			
		return false
	end

	diff = res1 - res2
	if equalIsBetter
		result = true if diff < -@eps
		result = "tie" if diff >= -@eps and diff <= @eps
		result = false if diff > @eps
	else 
		result = true if diff < -@eps
		result = false if diff >= -@eps
	end
#	p res1 - res2
#	p result
	return result
end

# =========================================================================================
# State1 dominates state2 iff it has at least as much detail and is better with the lesser detail.
# =========================================================================================
def dominates(state1_int, state2_int, equalIsBetter=true)
	detailBetter = (detail(state1_int) >= detail(state2_int))
	return false unless detailBetter
	return isBetterWithLesserDetail(state1_int, state2_int, equalIsBetter) 
end


# =========================================================================================
# Do runs for a state with a lower level of detail than some other one, up to a total bound on performance (from that other one).
# How that bound influences censoring thresholds depends on the objective function:
# 1) Median: all runs are cut off just after the bound, can stop after 50% cutoffs (if the bound exists, otherwise need a tie breaking criterion).
# 2) Mean, mean10, mean1000, etc: Accumulate time t_new one run at a time, counting timeouts appropriately. The timeout for each run is min(bound*numRuns-t_new, regular censoring time at this level)
# 3) Other objectives: not implemented yet, didn't really think about ways to exploit the bound there.
# =========================================================================================
def doRunsBounded(full_state_int, level, paramBound=nil)
	stripped_state_int = $full_int_to_stripped_int[full_state_int]
	
	bound = paramBound
	bound = nil if isPruned(paramBound)
		
	raise "Should not recompute runs for level #{level} that we already did and that weren't pruned: #{@cachedResultScalars[stripped_state_int][level]} for state #{stripped_state_int}." if @cachedResultScalars[stripped_state_int][level] and not isPruned(@cachedResultScalars[stripped_state_int][level])

	#=== Get the performance of the incumbent --- if that doesn't exist, get it first.
	incumbent_bound = @cachedResultScalars[$full_int_to_stripped_int[@incumbent_state_int]][level]
	if incumbent_bound == nil and not full_state_int == @incumbent_state_int
		if @cachedResultScalars[$full_int_to_stripped_int[@incumbent_state_int]][level-1] == nil
			raise "incumbent state has performance nil at level #{level} and nil at level #{level-1}: #{@cachedResultScalars[$full_int_to_stripped_int[@incumbent_state_int]].join("\n")}" 
		end
		
		output "State wants more detail (#{level}+1) than incumbent (#{detail(@incumbent_state_int)}), doing incumbent first:\n#{str(full_state_int)}\n#{str(@incumbent_state_int)}"
		boundedIncTimeSpentInState(@incumbent_state_int, nil)
		incumbent_bound = @cachedResultScalars[$full_int_to_stripped_int[@incumbent_state_int]][level]
	end

	#=== This heuristic changes the trajectory but makes everything a lot faster: limit configurations by @boundMultiplier*bound from incumbent.
	if @boundMultiplier and not full_state_int == @incumbent_state_int
		raise "Incumbent still doesn't has a bound for level #{level}" if incumbent_bound == nil

		#=== For FocusedILS, the incumbent may actually have been pruned at that level -- in that case, leave the bound at whatever it is.
		unless isPruned(incumbent_bound)
			bound_from_incumbent = @boundMultiplier *  incumbent_bound
			bound = bound_from_incumbent unless bound
			bound = [bound, bound_from_incumbent].min
		end
	end

#	p "doRunsBounded, level #{level}, bound #{bound} (parambound #{paramBound})"

	if level == @maxDetail
		output "Already at maximal level of detail #{@maxDetail} for state #{full_state_int}"
		return
	end

	numRuns = @numRunsPerLevelOfDetail[level]
	censorTime = @censoringThresholdPerLevelOfDetail[level]

#	puts "level=#{level}, bound=#{bound}, numRuns=#{numRuns}, censorTime=#{censorTime}"

	unless @cachedResultScalars.key?(stripped_state_int)
		@cachedResultScalars[stripped_state_int] = []
	end

	raise "doRunsBounded called with too low a level of detail - #{level} has been run for state #{stripped_state_int} already, result = #{@cachedResultScalars[stripped_state_int][level]}" if @cachedResultScalars[stripped_state_int].length > level and not isPruned(@cachedResultScalars[stripped_state_int][level])
	raise "doRunsBounded called with too high a level of detail - the highest level so far is #{@cachedResultScalars[stripped_state_int].length}, now supposed to do #{level} for state #{stripped_state_int}." if @cachedResultScalars[stripped_state_int].length < level

	#=== Compute the scalarResult depending on whether there is a bound and on the objective function. If censored it will be the threshhold + 1
	if bound and @pruning
		if @overall_obj == "median" and bound < censorTime - @eps
			########### This only works for the following definition of median:
			# ================ Let x1, x2, ..., x_n be an ordered series; then the median is x_{ceil(n/2)}. I.e. for 1 value, it's x_1; for 2 values, it's still x1; for 3 values, it's x_2, for 4 values, it is also x_2, etc.			
			singleResults = []
			allRunsPerformed = false

			thisCensorTime = bound + 0.01
			thisCensorTime = thisCensorTime.ceil.to_i if @algo =~ /spear/ or @algo =~ /sat4j/ or @algo =~ /smt/# Some algos can only handle full integers as cutoff.

			#=== We could be more pro-active here: it suffices to have >50% of the runs finish under the cutoff time.
			#=== So you can stop after that; but you only get an upper bound on your median and may have to rerun the rest later.
			#=== Since we want to keep it simple we don't do that here.
			
			#=== Just perform the runs with the target median as a bound.
			results = getRunsOnEntriesWithParams(@instances[0...numRuns], full_state_int, thisCensorTime)
			scalarResult = combinationOfObjectiveFunctions(@algo, @overall_obj, results, @run_obj, thisCensorTime, @cutoff_length).to_f
#			output "scalarResult #{scalarResult}, bound #{bound}"

			raise "Still need to implement @boundMultiplier for median" if @boundMultiplier
			#=== TODO: Make i the number of solved instances within the timelimit.
			i = 0
			scalarResult = ["pruned",i] if scalarResult > bound + @eps

		elsif @overall_obj =~ /mean/
			t_new = 0
			totalTimeBound = bound * numRuns
#			output " Bounded INCTIME: -> detail #{level}, N=#{numRuns}, censoring=#{censorTime}, totalTimeBound=#{totalTimeBound}"

			#=== Speed up the process by remembering some stuff from lower levels -- won't change trajectory.
			singleResults = []
			@savedResults = Hash.new unless @savedResults
			@savedResults[stripped_state_int] = Hash.new unless @savedResults.key?(stripped_state_int)
			unless level == 0
				if @savedResults[stripped_state_int][level-1].key?(censorTime)
					singleResults = @savedResults[stripped_state_int][level-1][censorTime].dup
					#=== Popping the last items from that list, since they might be pruned. One should have been enough, but there was still a difference in the trace - this may indicate something fishy, but it seems minor and not a priority.
					singleResults.pop
					singleResults.pop
				end
			end
			@savedResults[stripped_state_int][level] = Hash.new
			@savedResults[stripped_state_int][level][censorTime] = singleResults

			if singleResults == []
				t_new = 0
			else
				scalarResult = combinationOfObjectiveFunctions(@algo, @overall_obj, singleResults, @run_obj, censorTime, @cutoff_length).to_f
				t_new = scalarResult * (singleResults.length)
			end

			allRunsPerformed = true
			while singleResults.length < numRuns
				i = singleResults.length
				#=== Compute result and then put it into @cachedResultScalars.
				thisCensorTime = [[censorTime,0.01].max, (totalTimeBound-t_new) + 0.01].min 
				thisCensorTime = [[censorTime,0.01].max, (totalTimeBound-t_new).ceil.to_i].min if @algo =~ /spear/ or @algo =~ /sat4j/ or @algo =~ /smt/ # Some algos can only handle full integers as cutoff.

				if thisCensorTime <= @eps
					allRunsPerformed = false
					break 
				end
				singleResult = getRunsOnEntriesWithParams([@instances[i]], full_state_int, thisCensorTime)
				singleResults.concat singleResult
				
				scalarResult = combinationOfObjectiveFunctions(@algo, @overall_obj, singleResults, @run_obj, censorTime, @cutoff_length).to_f
				t_new = scalarResult * (i+1)
				
#				puts "   Used #{t_new}/#{totalTimeBound} seconds for this state, ran #{i+1}/#{numRuns} runs."
			end
			scalarResult = combinationOfObjectiveFunctions(@algo, @overall_obj, singleResults, @run_obj, censorTime, @cutoff_length).to_f

			#=== If the new configuration didn't have to do all runs, register it as pruned.
			if (scalarResult > bound+@eps and thisCensorTime < censorTime)
				scalarResult = ["pruned",i] 
			end
			scalarResult = ["pruned",i] unless allRunsPerformed
		else
			#=== Just perform the runs unboundedly.
			results = getRunsOnEntriesWithParams(@instances[0...numRuns], full_state_int, censorTime)
			scalarResult = combinationOfObjectiveFunctions(@algo, @overall_obj, results, @run_obj, censorTime, @cutoff_length).to_f
		end
	else
		#=== Just perform the runs unboundedly.
		results = getRunsOnEntriesWithParams(@instances[0...numRuns], full_state_int, censorTime)
		scalarResult = combinationOfObjectiveFunctions(@algo, @overall_obj, results, @run_obj, censorTime, @cutoff_length).to_f

		#=== Speed up the process by remembering some stuff from lower levels -- won't change trajectory.
		@savedResults = Hash.new unless @savedResults
		@savedResults[stripped_state_int] = Hash.new unless @savedResults.key?(stripped_state_int)
		@savedResults[stripped_state_int][level] = Hash.new
		@savedResults[stripped_state_int][level][censorTime] = results

	end
	@cachedResultScalars[stripped_state_int][level] = scalarResult

	#=== Care for the invariants to remain true:
	#=== 1) If the detail for this state is higher than the detail of the incumbent, catch up.
	if detail(@incumbent_state_int) < detail(full_state_int)
		output "State got more detail (#{detail(full_state_int)}) than incumbent (#{detail(@incumbent_state_int)}), catching up:\n#{str(full_state_int)}\n#{str(@incumbent_state_int)}"
		raise "This situation should not occur anymore now that the incumbent is always run first"
#			boundedIncTimeSpentInState(@incumbent_state_int, state)
		boundedIncTimeSpentInState(@incumbent_state_int, nil)
	end
	
	#=== 2) If this state is itself the incumbent one, output its new value.
	if full_state_int == @incumbent_state_int
		update_incumbent
		return
	end

	#=== 3) If the detail for this state and the incumbent is identical, compare the two -- in case of ties take the old one.
	if detail(@incumbent_state_int) == detail(full_state_int)
		highest_level = detail(full_state_int)
		inc_qual = @cachedResultScalars[$full_int_to_stripped_int[@incumbent_state_int]][highest_level-1]
		this_qual = @cachedResultScalars[stripped_state_int][highest_level-1]
#		output "inc_qual #{inc_qual}, this_qual #{this_qual}, highest level #{highest_level}" if @approach == "focused"
		output "bound for changing incumbent #{@nStart}" if @idn == 2
		unless shortBetter(inc_qual, this_qual, true)
			unless @idn == 2 and highest_level < @nStart
				output "New inc: #{this_qual}"

				raise "If two states have the same detail one must dominate the other: #{inc_qual}, #{this_qual}." unless shortBetter(this_qual, inc_qual, true)
				setNewIncumbent(full_state_int)
			end
		end
	end
end


# =========================================================================================
# Like boundedIncTimeSpentInState, but with bound on performance.
# =========================================================================================
def boundedIncTimeSpentInState(undetailedState_int, detailedState_int=nil)
	level = detail(undetailedState_int)
	return if level == @maxDetail
	
	bound = nil
	if detailedState_int
		raise "Detailed state must have larger detail than undetailed one: #{detail(detailedState_int)} > #{detail(undetailedState_int)} violates this" unless detail(detailedState_int) > detail(undetailedState_int)
		bound = @cachedResultScalars[$full_int_to_stripped_int[detailedState_int]][level]
		
		if isPruned(bound)
			doRunsBounded(detailedState_int, level, nil)
		end
		bound = @cachedResultScalars[$full_int_to_stripped_int[detailedState_int]][level]
	end
	doRunsBounded(undetailedState_int, level, bound)
end

# =========================================================================================
# Better: this is a completely symmetric comparison (except in the case of ties, in which case true is returned).
# At the end of this, one state must have at least as many evals as the other one and outperform it with the lower number of evals of the two.
# Cases:
# 1) state1 has more or equal number of evals and is better with those: return false (first is better)
# 2) state2 has more or equal number of evals and is better with less evals: return true (second is better)
# 3) state1 has less evals but is better with those: execute runs of state1 until either 1) or 2)
# 4) state2 has less evals but is better with those: execute runs of state2 until either 1) or 2)
# =========================================================================================
def better(new_state_int, old_state_int)
	visit(new_state_int)
	
	if @approach == "basic" or @approach == "random"
		raise "Old state must have been evaluated before" unless detail(old_state_int) == @maxDetail
		boundedIncTimeSpentInState(new_state_int, old_state_int)

		return true if dominates(new_state_int, old_state_int, false)
		return "tie" if dominates(new_state_int, old_state_int, true)
		return false if dominates(old_state_int, new_state_int, false) #=== The old state must really be better, ties don't count.
		return false if dominates(old_state_int, new_state_int, true) 
		raise "Still undecided at end of better between new state #{str(new_state_int)} and old state #{str(old_state_int)}. This shouldn't be possible."

	elsif @approach =~ /focused/
		#=== Everybody gets at least one more level of detail.
		if detail(new_state_int) < detail(old_state_int)
			boundedIncTimeSpentInState(new_state_int, old_state_int)
			@numEvaluationsThisIteration += 1
		elsif detail(new_state_int) > detail(old_state_int)
			boundedIncTimeSpentInState(old_state_int, new_state_int)
			@numEvaluationsThisIteration += 1
		else
			boundedIncTimeSpentInState(old_state_int, nil)
			@numEvaluationsThisIteration += 1

			if detail(new_state_int) < detail(old_state_int) # not always the case because if new_state is incumbent then we trigger a run of it.
				boundedIncTimeSpentInState(new_state_int, old_state_int)
				@numEvaluationsThisIteration += 1
			end
		end
		result = betterWithoutAutomaticIncrease(new_state_int, old_state_int)
	
		return result
	end
end

# =========================================================================================
# Helper function for better - add detail until one dominates.
# =========================================================================================
def betterWithoutAutomaticIncrease(new_state_int, old_state_int, haveToFinish=false)
	
	#=== If everything is clear, return.
	return true if dominates(new_state_int, old_state_int, false)
	return "tie" if dominates(new_state_int, old_state_int, true)
	return false if dominates(old_state_int, new_state_int, false)  #=== The old state must really be better, ties don't count.)
	return false if dominates(old_state_int, new_state_int, true) 
	
	#=== Otherwise, the one with less evaluations looks better.
	#Case 1: new one has less and is better.
	if detail(new_state_int) < detail(old_state_int)
		while not (dominates(new_state_int, old_state_int, true) or dominates(old_state_int, new_state_int, false))
			break if haveToStop() and not haveToFinish
			boundedIncTimeSpentInState(new_state_int, old_state_int)
		end
	end

	#Case 2: old one has less and is better.
	if detail(old_state_int) < detail(new_state_int)
		while not (dominates(new_state_int, old_state_int, true) or dominates(old_state_int, new_state_int, false))
			break if haveToStop() and not haveToFinish
			boundedIncTimeSpentInState(old_state_int, new_state_int)
		end
	end

	return true if dominates(new_state_int, old_state_int, false)
	return "tie" if dominates(new_state_int, old_state_int, true)
	return false if dominates(old_state_int, new_state_int, false)  #=== The old state must really be better, ties don't count.)

	#=== If we have to stop, don't move.
	return false if haveToStop() and not haveToFinish

	raise "Still undecided at end of better between new state #{str(new_state_int)} and old state #{str(old_state_int)}. This shouldn't be possible."
end
	

# =========================================
# Random search, for comparison.
# =========================================
def random_search(init_state_int)
	output "========================================================\nStarting RANDOM SEARCH \n========================================================"
	@iteration = 1
	current_state_int = init_state_int
	setNewIncumbent(current_state_int)
	visit(current_state_int)

	visited_states = Hash.new
	visited_states = [current_state_int]
	boundedIncTimeSpentInState(current_state_int, nil)

	@numEvaluationsThisIteration = 0
	if @iterativeDeepening and @init_def
		def_state_int = init_default()
		if better(def_state_int, current_state_int)
			output("          -> With this level of detail, the default is better than the result of the last phase, moving back to default.\n\n")  
			current_state_int = def_state_int
		else 
			output "        -> Result from last phase is better than default, keeping it.}"
		end
	end

	while not haveToStop()
		#=== Stop if all configurations have been evaluated -- otherwise, endless loop !
		while true
			random_state_int = init_random()
			next if visited_states.include?(random_state_int)
			visited_states << random_state_int
			break
		end
				
		if better(random_state_int, @incumbent_state_int)
#			checkIfNewIncumbent(random_state)
			output("          -> Take improving step to random #{str(random_state_int)}\n\n")  
		else 
#			checkIfNewIncumbent(current_state)
			output "        -> Worse random: #{str(random_state_int)}"
		end
		@iteration += 1
	end
end

# =========================================================================================
# =========================================================================================
#                  HERE THE ITERATED LOCAL SEARCH AND ITS SUBFUNCTIONS BEGIN.
# =========================================================================================
# =========================================================================================


# =========================================
# Stop search if one of the termination criteria is fulfilled.
# =========================================
def haveToStop()
	return true if $totalEvaluationCount>=@maxEvals
	return true if $total_cputime >= @tunerTimeout
	return true if @iteration > @maxIts
	return true if @algo =~ /saps/ and @allVisitedStates.size >2400
	#=== Also stop if we have already reached the minimum we can achieve
	return true if detail(@incumbent_state_int) == @maxDetail and @cachedResultScalars[$full_int_to_stripped_int[@incumbent_state_int]][detail(@incumbent_state_int)-1] < @optimum_solqual + @eps

	return false
end

def init_search
	#=== Init.
	if @init_def
		return init_default()
	else
		return  init_random()
	end
end


def iterative_deepening_ils(depth, init_state_int)
	@N = @nForDepth[depth-1]
	@cutoff_time = @cForDepth[depth-1]
	@tunerTimeout = @tForDepth[depth-1]
	@maxEvals = @eForDepth[depth-1]
	
	@iteration = 1
	@numFlip = 1

	@cachedResultScalars = Hash.new
	@numRunsPerLevelOfDetail = []
	@censoringThresholdPerLevelOfDetail = []

	puts("Level #{@depth}")
	@out.puts("Level #{@depth}")
#	@param_ils_traj_file.puts "Level #{@depth}"

	@allVisitedStates = {} # Hash indexed by state_string, containing number and all kinds of info.
	@orderedVisitedStates = []
	@statesVisited = []

	#=== Set up the schedule how many instances and which censoring thresholds to use at each level of detail.
	if @approach == "focused"
		@maxDetail = @N
		for i in 0...@maxDetail
			@numRunsPerLevelOfDetail[i] = i+1
			@censoringThresholdPerLevelOfDetail[i] = @cutoff_time
		end
		
	elsif @approach == "basic" or @approach == "random"
		@maxDetail = 1
		@numRunsPerLevelOfDetail[0] = @N
		@censoringThresholdPerLevelOfDetail[0] = @cutoff_time
	else
		raise "Unknown approach #{@approach}"
	end

	#=== Start the search for this level of ParamILS.
	output "========================================================\nStarting ILS for level #{depth}, i.e. a limit of N=#{@N}, and cutoff time=#{@cutoff_time}.\nCurrent CPU time = #{$total_cputime}, this run goes until #{@tunerTimeout} \n========================================================"
	if @approach == "random"
		random_search(init_state_int)
	else
		iterated_local_search(init_state_int)
	end
	output "Final solution for depth #{depth} with limit N=#{@N}, and cutoff time=#{@cutoff_time}."
	output_incumbent_state
	return @incumbent_state_int
end


# =========================================
# Standard ILS.
# =========================================
def iterated_local_search(init_state_int)
	current_state_int = init_state_int
	setNewIncumbent(current_state_int)
	visit(current_state_int)

	if @idn == 1
		@nStart.times{|i| boundedIncTimeSpentInState(current_state_int, nil)}
	else
		boundedIncTimeSpentInState(current_state_int, nil)
	end

	@numEvaluationsThisIteration = 0
	if @iterativeDeepening and @init_def
		def_state_int = init_default()
		if better(def_state_int, current_state_int)
			output("          -> With this level of detail, the default is better than the result of the last phase, moving back to default.\n\n")  
			current_state_int = def_state_int
		else 
			output "        -> Result from last phase is better than default, keeping it.}"
		end
	end

	#=== Perform @R random steps.
	@numEvaluationsThisIteration = 0
	for i in 0...@R
		break if haveToStop()
		random_state_int = init_random()
		
		@numEvaluationsThisIteration = 0
		if better(random_state_int, current_state_int)
			output("          -> Take improving step to random #{str(random_state_int)}\n\n")  
			current_state_int = random_state_int
		else 
			output "        -> Worse random: #{str(random_state_int)}"
		end
	end
	
	@numEvaluationsThisIteration=0
	last_ils_state_int = basic_local_search(current_state_int)
	current_state_int = last_ils_state_int
  
	@iteration += 1
	while not haveToStop()
		output("#{$totalEvaluationCount}/#{@maxEvals}, #{$total_cputime}/#{@tunerTimeout}")

                puts "iteration #{@iteration}, flip #{@numFlip}, evaluation count #{$totalEvaluationCount}"
                @out.puts "iteration #{@iteration}, flip #{@numFlip}, evaluation count #{$totalEvaluationCount}"

		current_state_int = perturbation(current_state_int, @pertubation_strength)
		visit(current_state_int)
		boundedIncTimeSpentInState(current_state_int, nil)

		current_state_int = basic_local_search(current_state_int)
		
		unless @pert_rand
			#=== Acceptance criterion.
			acc_state_int = acceptance_criterion(last_ils_state_int, current_state_int)
			#=== With low probability, random restart.
			if rand(100) < @p_restart*100
				output "Random re-initialisation to #{str(acc_state_int)}"
				acc_state_int = init_random() 
				visit(acc_state_int)
				boundedIncTimeSpentInState(acc_state_int, nil)
			end
			current_state_int = acc_state_int
		end
		last_ils_state_int = current_state_int
		@iteration += 1
	end
end

# =========================================
# Init local search with default.
# =========================================
def init_default()
	init_state = Hash.new
	for param in @params
		init_state[param] = @default[param]
	end
	for key in @start_ass.keys
		init_state[key] = @start_ass[key]
	end
	for key in @fixed_ass.keys
		init_state[key] = @fixed_ass[key]
	end
	return full_state_to_full_int(init_state)
end

# =========================================
# Init local search with random values whose combination is allowed.
# =========================================
def init_random()
	is_forbidden = true
	while is_forbidden
		init_state = Hash.new
		for param in @params
			init_state[param] = @domain[param][rand(@domain[param].length)]
		end
		for key in @fixed_ass.keys
			init_state[key] = @fixed_ass[key]
		end
		is_forbidden = forbidden(init_state, @forbidden_combos)
	end
	return full_state_to_full_int(init_state)
end

# =========================================
# Basic local search.
# =========================================
def basic_local_search(start_state_int)
	current_state_int = start_state_int
	
#	@numEvaluationsThisIteration = 0
	output "   BLS in iteration #{@iteration}, start with #{str(current_state_int)}\n"
	visited_states = [current_state_int]
	changed = true
	while changed and not haveToStop() # local search.
		changed = false
		#=== Get array of [neighbouring state, param to change, old value, value to set it to]
		neighbours = neighbourhood(current_state_int)
		numNeighboursEvaluated = 0
		cpuBefore = $total_cputime
		numEvalsBefore = $totalEvaluationCount
		#=== Move to the first better neighbour
		while neighbours.length > 0 and not haveToStop()
			rand_index = rand(neighbours.length)
			
			neighbour_full_state, param, oldVal, newVal = neighbours[rand_index]
			neighbour_int = full_state_to_full_int(neighbour_full_state)
			neighbours.delete_at(rand_index)

			#=== To avoid loops, disallow state we have been at in this iteration.
			next if visited_states.include?(neighbour_int) 
			visited_states << neighbour_int 

			output("    Changing #{changed_between_states(current_state_int, neighbour_int)}, evaluating ...")
			if better(neighbour_int, current_state_int)
				#=== TODO: check if tie, increment counters accordingly. (and change output)
				output("          -> Take improving step to neighbour #{str(neighbour_int)} with flip #{@numFlip}\n\n")  
				current_state_int = neighbour_int
				
				#=== Bonus runs in FocusedILS.
				if @approach =~ /focused/
					numBonus = @numEvaluationsThisIteration
					output("          \n============= Performing #{numBonus} bonus runs of state: #{str(current_state_int)} ============ \n\n")  
					addNewRuns(current_state_int, numBonus)
					output("          -> After #{numBonus} bonus runs: #{str(current_state_int)}\n\n")  
					@numEvaluationsThisIteration=0
				end

				##=== Treat this state as a local minimum if we have reached a maximum level of detail for the generation of LMs (maximal path length).
				#break if detail(state_string(current_state)) == @maxDetail

				changed = true
				break
			else 
				#=== TODO: check if tie, increment counters accordingly. (and change output)
				output "        -> worse: #{str(neighbour_int,false)}"
			end
			numNeighboursEvaluated += 1
			break if numNeighboursEvaluated > @numNeighboursToEvaluate
		end
		incumbent_bound = @cachedResultScalars[$full_int_to_stripped_int[@incumbent_state_int]][detail(@incumbent_state_int)-1]
		@stepstats_out.puts "#{@numFlip}, #{@iteration}, #{numNeighboursEvaluated+1}, #{$total_cputime-cpuBefore}, #{$totalEvaluationCount-numEvalsBefore}, #{incumbent_bound}"
		@stepstats_out.flush
		@numFlip += 1
	end
	
	#=== We now found a local minimum (w.r.t. the limited random neighbourhood we looked at)
	lm_state_int = current_state_int
	
	#=== Bonus runs in FocusedILS
	if @approach =~ /focused/
		numBonus = @numEvaluationsThisIteration
		output("          \n============= Performing #{numBonus} bonus runs of state: #{str(lm_state_int)} ============ \n\n")  
		addNewRuns(lm_state_int, numBonus)
		output("          -> After #{numBonus} bonus runs for LM: #{str(lm_state_int)}\n\n")  
		@numEvaluationsThisIteration=0
	end
	
	@allVisitedStates[lm_state_int]["timesLM"] += 1

	output "   LM for iteration #{@iteration}: #{str(lm_state_int)}\n"
	output "\n========== DETAILED RESULTS (iteration #{@iteration}): =========="
	output_result(lm_state_int)
	output "================================================\n"
	
	#=== Compare every LM against incumbent state.
	output "Comparing LM against incumbent:\n#{str(lm_state_int)}\n#{str(@incumbent_state_int)}"
	if betterWithoutAutomaticIncrease(lm_state_int, @incumbent_state_int, true)
		output "LM better, change incumbent"
		setNewIncumbent(lm_state_int)
	else
		output "Incumbent better, keeping it"
	end
	
	return lm_state_int
end

# =========================================
# Simple pertubation.
# =========================================
def perturbation(state_int, strength)
	if @pert_rand
		result_state_int = init_random
	else
		result_state_int = state_int
		strength.times{|x|
			neighbours = neighbourhood(result_state_int)
			rand_index = rand(neighbours.length)
			result_state_int = full_state_to_full_int(neighbours[rand_index][0])
#			puts "    perturb to ---> #{str(result_state_int)}" # Don't need to evaluate at this point!
			output "    perturb to ---> #{str(result_state_int)}" # Don't need to evaluate at this point!
		}
	end
	return result_state_int
end

# =========================================
# Acceptance criterion BETTER.
# =========================================
def acceptance_criterion(last_ils_state_int, current_state_int)
	if last_ils_state_int == current_state_int
		output "same state as last ILS: #{str(last_ils_state_int)}"
		@accepted_last = "same"
		return current_state_int
	end

	if betterWithoutAutomaticIncrease(current_state_int, last_ils_state_int) 
		output "   Accepting new better local optimum: #{str(current_state_int)}"
		@accepted_last = "acc"
		return current_state_int
	end

	output "rejecting worse #{str(current_state_int)}, going back to #{str(last_ils_state_int)}"
	@accepted_last = "rej"
	return last_ils_state_int
end

# =========================================
# Return what changed between two states.
# =========================================
def changed_between_states(old_state_int, new_state_int)
	res = []
	old_state = $int_to_full_state[old_state_int]
	new_state = $int_to_full_state[new_state_int]
	for key in old_state.keys
		res << "#{key}: #{old_state[key]}->#{new_state[key]}" if old_state[key] != new_state[key]
	end
	return res
end

def output_help(out)
	puts ARGV.join(" --- ")
	
	out.puts "\n======================================================================================================"
	out.puts "PARAM_ILS version 2.1, Frank Hutter, 2007. Optimizes parameters of an algorithm on a set of instances."
	out.puts "For more details, see: \"Frank Hutter, Holger Hoos, and Thomas Stueztle:"
	out.puts  "Automatic Algorithm Configuration based on Local Search. In Proc. of AAAI-07\"."
	out.puts "Usage: ruby param_ils_2_1_run.rb, followed by the following optional arguments (defaults in [parentheses])"

	out.puts "General"
	out.puts "-------\n"
	out.puts "-algo <X>, where <X> calls (a wrapper of) the algorithm to be optimized. [#{@algo}]"
	out.puts "           This wrapper/algorithm has to follow the input/output format in README.TXT" 
	out.puts "-run_obj <X>, where <X> in {runtime, runlength, speedup} [#{@run_obj}] (or user-defined in file algo_specifics.rb)"
	out.puts "-overall_obj <X>, where <X> in {mean, adj_mean, median, q90, geomean} [#{@overall_obj}]\n"
	out.puts "-cutoff_time <X>, where <X> is a float [#{@cutoff_time}]"
	out.puts "-cutoff_length <X>, where <X> is an integer [#{@cutoff_length}]"
	out.puts "-N <X>, where X is an integer indicating the maximal number of samples obtained per config [#{@N}]."# (BasicILS(N) or FocusedILS with a bound on each N(theta)) [#{@N}]"
	out.puts "-numRun <X>, where X is an integer indicating the how manieth run of this tuning scenario this is[#{@numRun}]."
	out.puts "-instance_seed_file <X>, where X is a filename [#{@instance_seed_file}]."
	out.puts "-test-instance_seed_file <X>, where X is either empty or a filename. Either means use half the training set as validation. [#{@test_instance_seed_file}]."
	out.puts "-test <X>, where X is IN {0,1}. In combination with scenario, use test set instead of validation set."

	out.puts "\nWHEN SHOULD the tuner TERMINATE?"
	out.puts "--------------------------------"
	out.puts "-maxEvals <X>, where <X> is an integer defining the budget of single algorithm evaluations [#{@maxEvals}]"
	out.puts "-tunerTimeout <X>, where <X> is a float (seconds) [#{@tunerTimeout}]"
	
	out.puts "\nGeneral CONTROL:"
	out.puts "----------"
	out.puts "-fixed <X>, where <X> = \"<param1> <val1> <param2> <val2> ...\" and those values will be fixed throughout the tuning [\"#{@fix_input}\"]"
	out.puts "-output_level <X>, where <X> is an integer (higher = more output) [#{@output_level}]"

	out.puts "-db <X>, where <X> in {0,1} [#{@db}]. I use a MQSQL database to store algorithm runs." 
#	out.puts "         so I don't have to redo them if I redo an experiment. You are welcome to look "
#	out.puts "         at my scripts and adapt them to fit your purpose, but I will not support this feature."
#	out.puts "         When DB is used, <algo> means the name of the algorithm in the DB."
	
	out.puts "-jobserver <X>, where <X> in {0,1} [#{@jobserver}]. For big jobs, I distribute calls to several machines."
#	out.puts "              using jobserver.rb, written by Christian Bang. I will not support this feature, but "
#	out.puts "              you are welcome to look at the scripts I use to dispatch the jobs"

	out.puts "\nCONTROL OF PARAM_ILS:"
	out.puts "---------------------"
	out.puts "-approach <X>, where <X> in {random, basic, focused} [#{@approach}]"
	out.puts "-init <X>, where <X> is either in {0,1} (0=random, 1=default), or <X> = \"<param1> <val1> <param2> <val2> ...\" [#{@init_def}]"

	out.puts "\nExamples:\n"
#	out.puts "ruby ../scripts/param_ils_2_1_run.rb gls4mpe -run_obj approx -overall_obj mean -N 50 -numParamILSRuns 10 -maxEvals 20000 -instanceFile gridGH25GW25K2-noquals-train-1-to-50.txt -db 1 -cluster 2"
#	out.puts "ruby ../scripts/explore_new_focused_ils.rb saps median_runlength adj_mean 100 10 2000000 normal 10 200 sw-sat-saps-1M-to-2M-train-first100.txt params.txt 1 0 0 2"
#	out.puts "ruby ../scripts/explore_new_focused_ils.rb saps median_runlength adj_mean 100 10 2000000 random 10 200 sw-sat-saps-1M-to-2M-train-first100.txt params.txt 1 0 0 2"
#	out.puts "ruby ../scripts/param_ils_2_1_run.rb -algo saps -db 1 -cluster 2"

#	out.puts "ruby ../scripts/explore_new_focused_ils.rb sat4j median_runlength adj_mean 100 10 2000000 normal 1 100000 sw-sat4j-10000-to-11000-randomized-train-first100.txt params.txt 1 1 0 2"
#	out.puts "ruby ../scripts/explore_new_focused_ils.rb sat4j median_runlength adj_mean 1 10 2000000 normal 1 100000 sw-sat4j-10000-to-11000-randomized-train-first100.txt params.txt 1 1 0 2"
#	puts "ruby focused_ils.rb branin_function median 100 10 _ test 10 100 _ 0 1" 
end

	
# =========================================
#                  MAIN
# =========================================

# ==== Init.

@eps = 1e-10
$statsfile = "jobserver_paramils_stats_#{rand}.txt"

#===  Create outfile
t = Time.now
now= t.strftime("%Y-%m-%dat%I-%M%p")

# =========================================
# ParamILS defaults.
# =========================================
@approach = "focused"
@init_def = 1
@algo = "saps"
@outdir = ""
@numRun = ""
@instance_seed_file = ""
@test_instance_seed_file = ""

@run_obj = "runtime"
@overall_obj = "adj_mean"
validN = 1000


@N = 2000
@cutoff_time = 10
@cutoff_length = "max"
@maxEvals = 100000
@tunerTimeout = 18000
@maxIts = 20000

@db = 1
@jobserver = 0

@fix_input = ""
start_input = ""

@output_level = 50
@test = false

$fakedCensoringForExactSameResults = false

@R = 10
@pertubation_strength = 3
@relative_pertubation_strength = 0.2
@pertubation_strength_scaling = false
@p_restart = 0.01
@numNeighboursToEvaluate = 1000000
@pruning = true
@iterativeDeepening = false
@boundMultiplier = 2
@idn = false

$minimum_runtime = 0.1
#=== lambda_t is the ratio of time allocated to ILS executions with lower N and c.
lambda_t = 0.5

#=== lambda_N is the ratio of @N allocated to the second last ILS execution.
lambda_n = 0.5

#=== lambda_c is the ratio of @cutoff_time allocated to the second last ILS execution.
lambda_c = 0.5

@pert_rand = false
@optimum_solqual = -1000000000 # stop once we reach this !

lambda_str = ""

# =========================================
# Read in ParamILS-specific command line options.
# =========================================
0.step(ARGV.length-1, 2){|i|
	case ARGV[i]
	when "-instance_seed_file"
		@instance_seed_file = ARGV[i+1]
	when "-test_instance_seed_file"
		@test_instance_seed_file = ARGV[i+1]
	when "-test"
		@test = ARGV[i+1]
		@test = false if @test == "0" or @test == "false" or @test == 0
	when "-approach"
		@approach = ARGV[i+1]
	when "-numRun"
		@numRun = ARGV[i+1].to_i
	when "-outdir"
		@outdir = ARGV[i+1]

	when "-algo"
		@algo = ARGV[i+1]
	when "-run_obj"
		@run_obj = ARGV[i+1]
	when "-overall_obj"
		@overall_obj = ARGV[i+1]
	when "-mintime"
		$minimum_runtime = ARGV[i+1].to_f
		
	when "-cutoff_time"
		@cutoff_time = ARGV[i+1]
	when "-cutoff_length"
		@cutoff_length = ARGV[i+1]
	when "-maxEvals"
		tmp = ARGV[i+1]
		tmp = 100000000000000000 if tmp == "max"
		@maxEvals = tmp.to_i
	when "-maxIts"
		tmp = ARGV[i+1]
		tmp = 100000000000000000 if tmp == "max"
		@maxIts = tmp.to_i
	when "-tunerTimeout"
		tmp = ARGV[i+1]
		tmp = 100000000000000000 if tmp == "max"
		@tunerTimeout = tmp.to_i
	when "-pruning"
		@pruning = ARGV[i+1]
	when "-faked"
		$fakedCensoringForExactSameResults = ARGV[i+1]
	when "-validN"
		validN = ARGV[i+1].to_i
	when "-idn"
		@idn = ARGV[i+1].to_i
	when "-lambda_c"
		lambda_c = ARGV[i+1].to_f
		lambda_str += "-lambda_c#{lambda_c}"
	when "-lambda_t"
		lambda_t = ARGV[i+1].to_f
		lambda_str += "-lambda_t#{lambda_t}"
	when "-lambda_n"
		lambda_n = ARGV[i+1].to_f
		lambda_str += "-lambda_n#{lambda_n}"
			
#=== tuning parameter
	when "-ps"
		@pertubation_strength = ARGV[i+1].to_i
	when "-rps"
		@relative_pertubation_strength = ARGV[i+1].to_f
	when "-psr"
		@pertubation_strength_scaling = ARGV[i+1].to_i
	when "-rw_prob"
		@p_restart = ARGV[i+1].to_f
	when "-R"
		@R = ARGV[i+1].to_i
	when "-mN"
		@numNeighboursToEvaluate = ARGV[i+1].to_i
#	when "-maxDetail"
#		@maxDetail = ARGV[i+1].to_i
	when "-prand"
		@pert_rand = ARGV[i+1]
	when "-id"
		@iterativeDeepening =  ARGV[i+1].to_i
	when "-bm"
		@boundMultiplier =  ARGV[i+1].to_f

	when "-init"
		if ARGV[i+1].to_s == "0" or  ARGV[i+1].to_s == "1"
			@init_def  = ARGV[i+1].to_s
			@init_def = false if @init_def == "0"
		else
			#=== Start input values are at least partly user-defined, rest: default.
			@init_def  = ARGV[i+1]
			fileWithDef = ARGV[i+1]
			File.open(fileWithDef, "r"){|file| start_input = file.gets.chomp}
		end
	when "-N"
		@N = ARGV[i+1].to_i
		
		
	when "-db"
		@db = ARGV[i+1]
	when "-jobserver"
		@jobserver = ARGV[i+1].to_i

	when "-paramFile"
		$param_file = ARGV[i+1]
	when "-fix"
		@fix_input  = ARGV[i+1]

	when "-output_level"
		@output_level = ARGV[i+1].to_i

	when "-scenario"
		@scenario = ARGV[i+1]

		case @scenario
		when "SAT4J-SW"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 10
			@cutoff_length = "max"

			@tunerTimeout = 36000
#			@tunerTimeout = 3600

			@instance_seed_file = "sw-sat4j-5-to-10s-scrambled-first1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "sw-sat4j-5-to-10s-scrambled-last1000-det_algo/instance_list_0.txt" if @test

		when "SAT4J-SWorig"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = 5
			@cutoff_length = "max"

			@tunerTimeout = 36000

			@instance_seed_file = "sw-sat4j-1-to-2s-scrambled-first1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "sw-sat4j-1-to-2s-scrambled-last1000-det_algo/instance_list_0.txt" if @test

		when "SPEARswv"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800 # 18000

			@instance_seed_file = "SWV-scrambled-first302-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat/instance_list_0.txt" if @test

		when "SMT"
			@algo = "smt"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 600
			@cutoff_length = "max"

			@tunerTimeout = 864000

			@instance_seed_file = "SMT-all-strat/instance_list_#{@numRun}.txt"
#			@test_instance_seed_file = "SWV-scrambled-last302-strat/instance_list_0.txt" if @test

		when "SPEAR-bmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000

			@instance_seed_file = "BMC-scrambled-first64-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat/instance_list_0.txt" if @test

		when "SPEAR-cbmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000

			@instance_seed_file = "BMC-scrambled-newsplit-train-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-newsplit-test-strat/instance_list_0.txt" if @test

		when "SPEAR-obmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000

			@instance_seed_file = "BMC-scrambled-first64-minisattime-ordered/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat/instance_list_0.txt" if @test

		when "SPEARibm"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800 # 18000 #8640000

			@instance_seed_file = "IBM_40randomsets-scrambled-first382-feasible287-strat/instance_list_#{@numRun}.txt"
			#			@instance_seed_file = "IBM_40randomsets-scrambled-first382-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "IBM_40randomsets-scrambled-last383-strat/instance_list_0.txt" if @test

		when "SPEARibmall"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800 # 18000 #8640000

			@instance_seed_file = "IBM_40randomsets-scrambled-first382-strat/instance_list_#{@numRun}.txt"
			#			@instance_seed_file = "IBM_40randomsets-scrambled-first382-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "IBM_40randomsets-scrambled-last383-strat/instance_list_0.txt" if @test


		when "SAT4J-bmc"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean"
			
			@cutoff_time = 10 #300 
			@cutoff_length = "max"

			@tunerTimeout = 864000

			@instance_seed_file = "BMC-scrambled-first64-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat-det_algo/instance_list_0.txt" if @test

		when "SAT4J-swv"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean"
			
			@cutoff_time = 10 #300
			@cutoff_length = "max"

			@tunerTimeout = 864000

			@instance_seed_file = "SWV-scrambled-first302-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat-det_algo/instance_list_0.txt" if @test

		when "SPEAR-pjava"
			@algo = "spear1.2.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 100000

			@instance_seed_file = "PJAVA-scrambled-first38-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "PJAVA-scrambled-last38-strat/instance_list_0.txt" if @test

		when "CPLEX50"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 5
			@cutoff_length = "max"

			@tunerTimeout = 600

#			@instance_seed_file = "cats-regions-goods100-bids500-det_algo/instance_list_#{@numRun}.txt"
			@instance_seed_file = "CATS50-first1000-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "CATS50-last1000-strat-det_algo/instance_list_0.txt" if @test

	when "CPLEX100"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 5
			@cutoff_length = "max"

			@tunerTimeout = 18000

#			@instance_seed_file = "cats-regions-goods100-bids500-det_algo/instance_list_#{@numRun}.txt"
			$algodir = "/.autofs/csother/ubccsprojectarrow/hutter/ParamILS/cplex/"
			@instance_seed_file = "#{$algodir}CATS100-first1000-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "#{$algodir}CATS100-last1000-strat-det_algo/instance_list_0.txt" if @test
			@test_instance_seed_file = "#{$algodir}CATS100-first1000-strat-det_algo/instance_list_#{@numRun}.txt" unless @test  # same file for training and validation since I'm short on instances and CPLEX is deterministic.

		when "CPLEX200"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800

			@instance_seed_file = "CATS200-first1000-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "CATS200-last1000-strat-det_algo/instance_list_0.txt" if @test

		when "CPL-MIK"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800

			@instance_seed_file = "BCOL-mixed-integer-knapsack-scrambled-first60-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BCOL-mixed-integer-knapsack-scrambled-last60-strat-det_algo/instance_list_0.txt" if @test

		when "CPL-CLS"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800

			@instance_seed_file = "BCOL-capacitated_lot_sizing-scrambled-first50-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BCOL-capacitated_lot_sizing-scrambled-last50-strat-det_algo/instance_list_0.txt" if @test

		when "CPLCsch"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800

			@instance_seed_file = "BCOL-conic.sch-scrambled-first172-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BCOL-conic.sch-scrambled-last171-strat-det_algo/instance_list_0.txt" if @test


		when "CPLEX-miplib-old3"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 100
			@cutoff_length = "max"

			@tunerTimeout = 200000

			@instance_seed_file = "miplib3-scrambled-first33-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "miplib3-scrambled-last32-strat-det_algo/instance_list_0.txt" if @test
		
		when "CPLEX-miplib"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800 # 18000 #200000

			@instance_seed_file = "miplib2003_allsolved-scrambled-first26-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "miplib2003_allsolved-scrambled-last26-strat-det_algo/instance_list_0.txt" if @test

		
		when "SAPS-SW-median" #orig for the AAAI paper
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = 10
			@cutoff_length = "max"
			
#			@tunerTimeout = 36000
			@tunerTimeout = 3600

			@instance_seed_file = "SW-saps-5s-to-10s-scrambled-first113/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-5s-to-10s-scrambled-last100/instance_list_0.txt" if @test

		when "SAPS-SW"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 10
			@cutoff_length = "max"
			
#			@tunerTimeout = 36000
			@tunerTimeout = 3600

			@instance_seed_file = "SW-saps-5s-to-10s-scrambled-first113/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-5s-to-10s-scrambled-last100/instance_list_0.txt" if @test

		when "SAPS-SWorig"
			@algo = "saps"

			@run_obj = "runlength"
			@overall_obj = "median"

			@cutoff_time = "max"
			@cutoff_length = "2000000"
			
			@tunerTimeout = 36000

			@instance_seed_file = "SW-saps-1M-to2M-scrambled-first100/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-1M-to2M-scrambled-last82/instance_list_0.txt" if @test

		when "SAPS-SW100"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = "max"
			@cutoff_length = 2000000
			
			@tunerTimeout = 18000

			@instance_seed_file = "SW-saps-100s-to200s-scrambled-first231/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-100s-to200s-scrambled-last100/instance_list_0.txt" if @test

		when "SAPS-QWH"
			@algo = "saps"

			@run_obj = "runlength"
#			@overall_obj = "q90"
			@overall_obj = "median"

			@cutoff_time = 1 #0.5 #1
			@cutoff_length = "max"
			
			@tunerTimeout = 3600 #9600

			@instance_seed_file = "qwh_med_inst/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "qwh_med_inst_test/instance_list_0.txt" if @test

		when "SAPS-QWH-mean"
			@algo = "saps"

			@run_obj = "runlength"
#			@overall_obj = "q90"
			@overall_obj = "mean"

			@cutoff_time = 1 #0.5 #1
			@cutoff_length = "max"
			
			@tunerTimeout = 3600 #9600

			@instance_seed_file = "qwh_med_inst/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "qwh_med_inst_test/instance_list_0.txt" if @test

		when "SAPS-QWHorig"
			@algo = "saps"

			@run_obj = "runlength"
			@overall_obj = "q90"

			@cutoff_time = 1
			@cutoff_length = "max"
			
			@tunerTimeout = 9000

			@instance_seed_file = "qwh_med_inst/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "qwh_med_inst_test/instance_list_0.txt" if @test

		when "GLS4MPE"
			@algo = "gls4mpe"

			@run_obj = "prob"
			@overall_obj = "mean"

			@cutoff_time = 10
			@cutoff_length = "max"
			
			@tunerTimeout = 200000

			@instance_seed_file = "gridGH25GW25K2-pseudo-opt-scrambled-first50/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "gridGH25GW25K2-pseudo-opt-scrambled-last50/instance_list_0.txt" if @test

		when "SWa-Spea"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			$algodir = "/.autofs/csother/ubccsprojectarrow/hutter/ParamILS/spear/"
			@instance_seed_file = "#{$algodir}SWGCP-all-scrambled-first1000-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "#{$algodir}SWGCP-all-scrambled-last1000-strat/instance_list_0.txt" if @test
			@test_instance_seed_file = "#{$algodir}validation-SWGCP-late-all-scrambled-first1000-strat/instance_list_#{@numRun}.txt" unless @test

		when "QCP-Spea"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			$algodir = "/.autofs/csother/ubccsprojectarrow/hutter/ParamILS/spear/"
			@instance_seed_file = "#{$algodir}QCP-scrambled-first1000-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "#{$algodir}QCP-scrambled-last1000-strat/instance_list_0.txt" if @test
			@test_instance_seed_file = "#{$algodir}validation-QCP-scrambled-first1000-strat/instance_list_#{@numRun}.txt" unless @test

		when "Spear-SW"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 10
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			@instance_seed_file = "SWGCP-spear-5-to10s-scrambled-first600-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWGCP-spear-5-to10s-scrambled-last551-strat/instance_list_0.txt" if @test

		when "SWa-S4J"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			@instance_seed_file = "SWGCP-all-scrambled-first1000-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWGCP-all-scrambled-last1000-strat-det_algo/instance_list_0.txt" if @test

		when "SWa-SAPS-median"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			@instance_seed_file = "SWGCP-allSat-scrambled-first1000-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWGCP-allSat-scrambled-last1000-strat/instance_list_0.txt" if @test

		when "QCP-SAPS"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000
			
			$algodir = "/.autofs/csother/ubccsprojectarrow/hutter/ParamILS/saps-qcp/"
			@instance_seed_file = "#{$algodir}QCP-sat-scrambled-first1000-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "#{$algodir}QCP-sat-scrambled-last1000-strat/instance_list_0.txt" if @test
			@test_instance_seed_file = "#{$algodir}validation-QCP-sat-scrambled-first1000-strat/instance_list_#{@numRun}.txt" unless @test

		when "SWa-SAPS"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			$algodir = "/.autofs/csother/ubccsprojectarrow/hutter/ParamILS/saps-sw/"
			@instance_seed_file = "#{$algodir}SWGCP-allSat-scrambled-first1000-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "#{$algodir}SWGCP-allSat-scrambled-last1000-strat/instance_list_0.txt" if @test
			@test_instance_seed_file = "#{$algodir}validation-SWGCP-allSat-scrambled-first1000-strat/instance_list_#{@numRun}.txt" unless @test

		when "SAPSrand"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"
			
			@tunerTimeout = 172800

			@instance_seed_file = "SatComp03to07_sat-random-scrambled-first363-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SatComp03to07_sat-random-scrambled-last363-strat/instance_list_0.txt" if @test

		when "SAPScraf"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"
			
			@tunerTimeout = 172800

			@instance_seed_file = "SatComp03to07_sat-crafted-scrambled-first189-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SatComp03to07_sat-crafted-scrambled-last188-strat/instance_list_0.txt" if @test

		when "MD5"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 1000
			@cutoff_length = "max"
			
			@tunerTimeout = 180000

			@instance_seed_file = "MD5-21instances-strat/instance_list_#{@numRun}.txt"

		when "MD5-ordered"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 1000
			@cutoff_length = "max"
			
			@tunerTimeout = 180000

			@instance_seed_file = "MD5-21instances-strat-ordered/instance_list_#{@numRun}.txt"
			
		when "SWV-1"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 1
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			@instance_seed_file = "SWV-scrambled-first302-below1s-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-below1s-strat/instance_list_0.txt" if @test

		when "SWV-10"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 10
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			@instance_seed_file = "SWV-scrambled-first302-below10s-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-below10s-strat/instance_list_0.txt" if @test

		when "SWV-100"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 100
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			@instance_seed_file = "SWV-scrambled-first302-below100s-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-below100s-strat/instance_list_0.txt" if @test

		when "SWV-1000"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 1000
			@cutoff_length = "max"
			
			@tunerTimeout = 18000

			@instance_seed_file = "SWV-scrambled-first302-below1000s-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-below1000s-strat/instance_list_0.txt" if @test

		when "SWVc1"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean" #"adj_mean"
			
			@cutoff_time = 1
			@cutoff_length = "max"

			@tunerTimeout = 18000

			@instance_seed_file = "SWV-scrambled-first302-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat/instance_list_0.txt" if @test

		when "SWVc10"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean" #"adj_mean"
			
			@cutoff_time = 10
			@cutoff_length = "max"

			@tunerTimeout = 18000

			@instance_seed_file = "SWV-scrambled-first302-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat/instance_list_0.txt" if @test

		when "SWVc100"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean" #"adj_mean"
			
			@cutoff_time = 100
			@cutoff_length = "max"

			@tunerTimeout = 18000

			@instance_seed_file = "SWV-scrambled-first302-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat/instance_list_0.txt" if @test

		when "SWVc1000"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean" #"adj_mean"
			
			@cutoff_time = 1000
			@cutoff_length = "max"

			@tunerTimeout = 18000

			@instance_seed_file = "SWV-scrambled-first302-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat/instance_list_0.txt" if @test

		when "CPLEXOR"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 18000

			@instance_seed_file = "OrLibSax-scrambled-first70-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "OrLibSax-scrambled-last70-strat-det_algo/instance_list_0.txt" if @test

		when "CORlong"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800

			@instance_seed_file = "OrLibSax-scrambled-first70-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "OrLibSax-scrambled-last70-strat-det_algo/instance_list_0.txt" if @test

		when "CPLEXQP"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean"

			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 172800

			@instance_seed_file = "mirela_qp-scrambled-first1000-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "mirela_qp-scrambled-last1000-strat-det_algo/instance_list_0.txt" if @test

=begin
		when "metatune"
			@algo = "ruby param_ils_caller_for_metatuning.rb"
			@db = false

			@run_obj = "qual"
			@overall_obj = "mean"

			@cutoff_time = 100
			@cutoff_length = "max"

			@tunerTimeout = 400; #2592000 # one month

			@instance_seed_file = "metatune_scearios_train.txt"
			@test_instance_seed_file = "metatune_scearios_test.txt" if @test
=end
		when "metatune"
			@algo = "param_ils_metatune"
#			@db = false

			@run_obj = "qual"
			@overall_obj = "mean"

			@cutoff_time = 18000 #100
			@cutoff_length = "max"

			@tunerTimeout = 172800000 # 2000 days. 10000

#			@instance_seed_file = "metatune_scearios_train-4.txt"
			@instance_seed_file = "metatune_scearios_train-20each.txt"
			@test_instance_seed_file = "metatune_scearios_test-3.txt" if @test

		when "metatu2"
			@algo = "param_ils_metatune"
#			@db = false

			@run_obj = "qual"
			@overall_obj = "mean"

			@cutoff_time = 3601 #100
			@cutoff_length = "max"

			@tunerTimeout = 172800000 # 2000 days. 10000

#			@instance_seed_file = "metatune_scearios_train-4.txt"
			@instance_seed_file = "metatune_scearios_train-20each.txt"
			@test_instance_seed_file = "metatune_scearios_test-3.txt" if @test

		else
			raise "Undefined scenario #{@scenario}"
		end

	else
		output_help($stdout)
		puts "\n\n"
		raise "Unknown argument: #{ARGV[i]}"
	end
}

if ARGV.length < 1
	output_help($stdout)
	exit -1
end
Dir.chdir($algodir) if $algodir
$param_file = "#{$algodir}params.txt"

# =========================================
#=== Deal with parameters.
# =========================================
@maxDetail = @N
validN = 1000 unless @test

@db = false if @db == "0" or @db  == "false" or @db  == 0
@pruning = false if @pruning == "0" or @pruning  == "false" or @pruning  == 0
$fakedCensoringForExactSameResults = false if $fakedCensoringForExactSameResults == "0" or $fakedCensoringForExactSameResults  == "false" or $fakedCensoringForExactSameResults == 0
@iterativeDeepening = false if @iterativeDeepening == "0" or @iterativeDeepening  == "false" or @iterativeDeepening == 0
@idn = false if @idn == "0" or @idn == "false" or @idn == 0

@pertubation_strength_scaling = false if @pertubation_strength_scaling == "0" or @pertubation_strength_scaling == "false" or @pertubation_strength_scaling == 0

id_string = ""
id_string = "-id" if @iterativeDeepening
id_string = "-idn" if @idn == 1
id_string = "-idnn" if @idn == 2

raise "idn but no id" if @idn and not @iterativeDeepening 

@pert_rand = false if @pert_rand == "0" or @pert_rand == "false" or @pert_rand == 0
@pert_rand = true if @pert_rand == "1" or @pert_rand == "true" or @pert_rand == 1

pstring = ""
pstring = "prand-" if @pert_rand

#=== Only I am using the database.
$user_frank = @db
$fakedCensoringRuntime = @cutoff_time

fakedExactString = ""
fakedExactString = "faked-" if $fakedCensoringForExactSameResults

pruning_string = ""
pruning_string = "-prune" if @pruning
	
@cutoff_time = 1000000000 if @cutoff_time == "max"
@cutoff_time = @cutoff_time.to_f

@cutoff_length = 2147483647 if @cutoff_length == "max"
@cutoff_length = @cutoff_length.to_i

@maxEvals = 1000000000 if @maxEvals == "max"

@oncluster = @jobserver

if @instance_seed_file == ""
	output_help($stdout)
	puts "\n\n"
	raise "Need to specify -instance_list <instance_list>" 
end

if @numRun == ""
	output_help($stdout)
	puts "\n\n"
	raise "Need to specify #{-numRun} <numRun>" 
end

if @outdir == ""
	@outdir, tmp = File.split(@instance_seed_file)
end

# =========================================
# Initial setup.
# =========================================
require "dbi_ils_accessor.rb"  if @db
$total_cputime = 0

#=== Read params from param file.
@params, @domain, @default, $conditionals, @forbidden_combos = read_params($param_file) # Need conditionals as global to enable strip_state in param_reader
num_params = 0
num_combos = 1
for param in @params
	num_params += 1 if @domain[param].length>1
	num_combos *= @domain[param].length
end
puts "num_params = #{num_params}, num_combos=#{num_combos}"

if @pertubation_strength_scaling
	pert_string = "psr#{@relative_pertubation_strength}-"
	@pertubation_strength = [2, (@relative_pertubation_strength*num_params).ceil].max
else
	pert_string = "ps#{@pertubation_strength}-"
end

bm_string = ""
bm_string = "-bm#{@boundMultiplier}" if @boundMultiplier

mNString = ""
mNString = "#-mN{@numNeighboursToEvaluate}" unless @numNeighboursToEvaluate == 1000000

meta_string = ""
meta_string = "-metatune" unless @test

seed_string = "newsed2"

max_it_string = ""
max_it_string = "-maxit#{@maxIts}" unless @maxIts == 20000

log_filename = "#{@outdir}/#{@approach}-#{fakedExactString}#{pstring}#{pert_string}log-algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-time#{@cutoff_time}#{mNString}-init#{@init_def}#{id_string}#{seed_string}#{pruning_string}#{bm_string}#{lambda_str}#{max_it_string}_#{@numRun}#{meta_string}.txt"
traj_filename = "#{@outdir}/#{@approach}-#{fakedExactString}#{pstring}#{pert_string}traj-algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-time#{@cutoff_time}#{mNString}-init#{@init_def}#{id_string}#{seed_string}#{pruning_string}#{bm_string}#{lambda_str}#{max_it_string}_#{@numRun}#{meta_string}.txt"
stats_filename = "#{@outdir}/#{@approach}-#{fakedExactString}#{pstring}#{pert_string}stats-algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-time#{@cutoff_time}#{mNString}-init#{@init_def}#{id_string}#{seed_string}#{pruning_string}#{bm_string}#{lambda_str}#{max_it_string}_#{@numRun}#{meta_string}.txt"
stepstats_filename = "#{@outdir}/#{@approach}-#{fakedExactString}#{pstring}#{pert_string}stepstats-algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-time#{@cutoff_time}#{mNString}-init#{@init_def}#{id_string}#{seed_string}#{pruning_string}#{bm_string}#{lambda_str}#{max_it_string}_#{@numRun}#{meta_string}.txt"
test_output_filename = "#{@outdir}/#{@approach}-#{fakedExactString}#{pstring}#{pert_string}algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-time#{@cutoff_time}#{mNString}-init#{@init_def}#{id_string}#{seed_string}#{pruning_string}#{bm_string}#{lambda_str}#{max_it_string}_#{@numRun}#{meta_string}_test.txt"


@out = File.open(log_filename, "w")
@param_ils_traj_file = File.open(traj_filename, "w")
@stepstats_out = File.open(stepstats_filename , "w")

string_arguments = ARGV.map{|x| "\"#{x}\""}.join(" ")
#string_arguments = ARGV.join(" ")
@out.puts "Call: /usr/bin/ruby ../scripts/param_ils_2_1_run.rb #{string_arguments}\n\n\n"
#output_help(@out)
@out.flush

#=== Build fixed parameters.
@fixed_ass = set_fixed_params(@fix_input, @domain, @params, @out)
@start_ass = set_fixed_params(start_input, @domain, @params, @out)

@number_of_params_to_opt = @params.length - @fixed_ass.keys.length

#=== Build @instances from file
@instances = []
File.open(@instance_seed_file){|file| 
    while line=file.gets;
	raise "Incorrect number of entries #{line.strip.split.length} in #{line}" unless line.strip.split.length >= 3  
	entries = line.strip.split
	seed = entries[0].to_i
	inst = entries[1]
	desired_qual = entries[2]
	reference = entries[3]
	@instances << {"seed"=>seed, "name"=>inst, "qual"=>desired_qual, "reference"=>reference, "resultForState"=>Hash.new}
    end
}

# =========================================
#=== Split into training and test.
# =========================================
@test_instances = []
if @test_instance_seed_file == ""
	#=== N for both train and "test"
	@test_instances = @instances#[@N...@instances.length] -  use all training instances for validation
	#validN = 100
else 
	#=== N new ones for test
	@test_instances = []
	File.open(@test_instance_seed_file){|file| 
	    while line=file.gets;
		raise "Incorrect number of entries #{line.strip.split.length} in #{line}" unless line.strip.split.length >= 3  
		entries = line.strip.split
		seed = entries[0].to_i
		inst = entries[1]
		desired_qual = entries[2]
		reference = entries[3]
		@test_instances << {"seed"=>seed, "name"=>inst, "qual"=>desired_qual, "reference"=>reference, "resultForState"=>Hash.new}
	    end
	   # validN = 100
	}
end
@instances = @instances[0...@N]
@test_instances = @test_instances[0...validN]


# =========================================
#=== Output parameters.
# =========================================
seed = (@numRun+1)%100*1234 # Seed for Param_ILS, not the one to pass on to algorithms.
puts "seed: #{seed}"
@out.puts "seed: #{seed}"

@out.puts "algo: #{@algo}"
@out.puts "tunerTimeout: #{@tunerTimeout}"
@out.puts "maxEvals: #{@maxEvals}"

@out.puts "run_obj: #{@run_obj}"
@out.puts "overall_obj: #{@overall_obj}"

@out.puts "instance_seed_file: #{@instance_seed_file}"
@out.puts "test_instance_seed_file: #{@test_instance_seed_file}" unless @test_instance_seed_file == ""
@out.puts "N: #{@N}"

@out.puts "cutoff_time: #{@cutoff_time}"
@out.puts "cutoff_length: #{@cutoff_length}"

@out.puts "R: #{@R}"
@out.puts "pertubation_strength_basic: #{@pertubation_strength_basic}"
@out.puts "pertubation_strength_scaling: #{@pertubation_strength_scaling}"
@out.puts "p_restart: #{@p_restart}"

srand(seed)

# =========================================
#=== Execute ParamILS.
# =========================================

$total_cputime = 0
$totalEvaluationCount = 0
@lastOutputCPUTime = 0

puts("Run #{@numRun+1}")
@out.puts("Run #{@numRun+1}")
@param_ils_traj_file.puts "Run #{@numRun+1}"
init_state_int = init_search()

if @iterativeDeepening
	if lambda_n < 1-@eps
		numDepths = (Math.log(@N) / Math.log(1.0/lambda_n)).ceil + 1
	else
		numDepths = (Math.log(@cutoff_time) / Math.log(1.0/lambda_c)).ceil + 1
		p numDepths 
	end

	@nForDepth = []
	@cForDepth = []
	@tForDepth = []
	@eForDepth = []

	numDepths.times{|i| 
		@nForDepth[i] = ((@N+0.0) * (lambda_n**(numDepths-1-i))).ceil
		@cForDepth[i] = ((@cutoff_time+0.0) * (lambda_c**(numDepths-1-i))).ceil
		@tForDepth[i] = (@tunerTimeout+0.0) * (lambda_t**(numDepths-1-i))
		@eForDepth[i] = @maxEvals # (@maxEvals+0.0) * (lambda_t**(numDepths-1-i))
		puts "#{@nForDepth[i]} #{@cForDepth[i]} #{@tForDepth[i]}"
	}

	@nStart = 1
	numDepths.times{|depth|
		init_state_int = iterative_deepening_ils(depth+1, init_state_int)
		@R = 0 #=== Don't need random moves anymore.
		
		#=== For FocusedILS: Compute number of runs to give the incumbent from the last phase to start the new phase.
		if @idn and depth != numDepths-1
			@nStart = (detail(init_state_int) * @cForDepth[depth]) / @cForDepth[depth+1]
		else
			@nStart = 1
		end
	}
else
	@nStart = 1
	@nForDepth = [@N]
	@cForDepth = [@cutoff_time]
	@tForDepth = [@tunerTimeout]
	@eForDepth = [@maxEvals]
	iterative_deepening_ils(1, init_state_int)
end

# =========================================
#=== Statistics after the execution is over.
# =========================================

#=== Output gathered statistics.
puts stats_filename
File.open(stats_filename, "w"){|file|
	for full_state_int in @orderedVisitedStates
		hash = @allVisitedStates[full_state_int]
		resDetail = detail(full_state_int)
		cens = @censoringThresholdPerLevelOfDetail[resDetail-1]
		n = @numRunsPerLevelOfDetail[resDetail-1]
		res = @cachedResultScalars[$full_int_to_stripped_int[full_state_int]][resDetail-1]

		file.puts "#{hash["number"]}, #{res}, #{n}, #{hash["iteration"]}, #{cens}, #{hash["str"]}"	
#		file.puts "#{hash["number"]}, #{resDetail}, #{res}, #{hash["numVisited"]}, #{hash["totalTime"]}, #{hash["timesLM"]}, #{hash["timesMadeIncumbent"]}"
	end
}

#=== Evaluate test/validation performance.
output "\nComputing validation result ..."

test_results = getRunsOnEntriesWithParams(@test_instances, @incumbent_state_int, @cutoff_time, true)
for i in 0...@test_instances.length
	output "#{@test_instances[i]}: #{test_results[i]}"
end
test_result = combinationOfObjectiveFunctions(@algo, @overall_obj, test_results, @run_obj, @cutoff_time, @cutoff_length)
output "Combined result: #{test_result}"
File.open(test_output_filename, "w"){|file|
	file.puts test_result
}

@out.close
@param_ils_traj_file.close
@stepstats_out.close

puts "Validation result: #{test_result}"