# Summarizes the performance of a number of runs.

def round(x)
	a = 100.0
	return ((x * a)+0.5).floor/a
end

def output_help(out)
	out.puts "ruby perf_summary.rb <file name prefix with dir>"
	out.puts "Example:\n"
	out.puts "ruby ../scripts/perf_summary.rb qwh_med_inst/focused-algosaps-runobjrunlength-overallobjq90-runs100-maxEval20000-init1-time1.0-length2147483647"
	out.puts "ruby ../scripts/perf_summary.rb SW-saps-100s-to200s-scrambled-first231/focused-algosaps-runobjruntime-overallobjmedian-runs100-maxEval20000-init1-time1000000000.0-length2000000"
	out.puts "ruby ../scripts/perf_summary.rb gridGH25GW25K2-pseudo-opt-scrambled-first50/focused-algogls4mpe-runobjprob-overallobjmean-runs100-maxEval20000-init1-time10.0-length2147483647"
	out.puts "ruby ../scripts/perf_summary.rb sw-sat4j-1s-to-2s-random-train-first1000-scrambled-det_algo/focused-algosat4j-runobjruntime-overallobjmedian-runs100-maxEval20000-init1-time5.0-length2147483647"
	out.puts "ruby ../scripts/perf_summary.rb cats-regions-goods100-bids500-det_algo/basic-algocplex-runobjruntime-overallobjadj_mean-runs100-maxEval20000-init1-time10.0-length2147483647"
end

if ARGV.length < 1
	output_help($stdout)
	exit
end

# =========================================
# Read in command line options.
# =========================================
wholeprefix = ARGV[0]
dir, prefix = File.split(wholeprefix)
#modprefix = prefix.sub(/-algo/,"-traj-algo")
modprefix = prefix
prefix = prefix.sub(/-traj-algo/,"-algo")


trainresults = []
testresults = []
numIterations = []
ns = []
cutoff_times = []

trainfiles = []
testfiles = []

minTrain = 1e20
minInd = nil

Dir.entries(dir).sort.map{|entry|
	if entry =~ /^#{modprefix}/ and File.size?(dir + "/" + entry)
		File.open(dir + "/" + entry){|file|
			while line=file.gets
				savedline = line
			end
			trainresult = savedline.split[1].to_f
			trainresults << trainresult
			
			if trainresult < minTrain
				minTrain = trainresult
				minInd = trainresults.length - 1
			end
			
			ns << savedline.split[2].to_i
			numIterations << (savedline.split[3].to_i  - 1)
			cutoff_times << savedline.split[4].to_f
		}
		trainfiles << entry
		
		testfile = dir + "/" + entry.sub(/-traj-algo/,"-algo").sub(/\.txt/,"_test.txt")
#		p testfile
		if File.file?(testfile)
			File.open(testfile){|file|
				testresults << file.gets.to_f
			}
			testfiles << testfile.to_s
		end
	end
}

puts "#{trainfiles.length} Train files:\n#{trainfiles.join("\n")}\n\n"
puts "#{testfiles.length} Test files:\n#{testfiles.join("\n")}\n\n"

puts "Train results: #{trainresults.join(", ")}"
puts "Test results: #{testresults.join(", ")}"

#puts
require "stats_ils.rb"
puts "Training: #{trainresults.avg} +/- #{trainresults.std}, $#{round(trainresults.avg)} \\pm #{round(trainresults.std)}$" # \\pm #{ns.std}$  \\pm #{numIterations.std}
puts "Test: #{testresults.avg} +/- #{testresults.std}, $#{round(testresults.avg)} \\pm #{round(testresults.std)}$ & "
cc = correlation_coefficient(trainresults,testresults)
testOfBestTrain = testresults[minInd]
#puts
puts "Best train result: #{trainresults[minInd]}, test result of that: #{testOfBestTrain}"
puts "Correlation train/test: #{cc}"
#puts
puts "$#{round(trainresults.avg)} \\pm #{round(trainresults.std)}$ & "
puts "$#{round(testresults.avg)} \\pm #{round(testresults.std)}$ & " # (#{round(testOfBestTrain)})$ & "
puts "$#{ns.avg}$, $#{numIterations.avg}$ & " #, $#{round(cc)}$ & "
