require "param_reader.rb"
#require "global_helper.rb" # for instance hash


# =========================================================================================
# Output incumbent state is found.
# =========================================================================================
def output(state)
	#=== Fill in the unnecessary parameters with their default values for output.
	filled_state = state.dup
	for param in @params
		if state.key?(param)
			filled_state[param] = state[param]
		else
			filled_state[param] = @default[param]
			filled_state[param] = @fixed_ass[param] if @fixed_ass.key?(param)
		end
	end

	val = -1
	numeval = 0
	$total_cputime = 0
	
	@out.print "#{$total_cputime}, #{val}, #{numeval}, 0, 0, "
	@out.puts @params.map{|x|filled_state[x]}.join(", ") # output the current values
	@out.flush
end


# =========================================
# Init local search with random values whose combination is allowed.
# =========================================
def init_random()
	is_forbidden = true
	while is_forbidden
		init_state = Hash.new
		for param in @params
			init_state[param] = @domain[param][rand(@domain[param].length)]
		end
		is_forbidden = forbidden(init_state, @forbidden_combos)
	end
	return init_state
end


def output_help(out)
	out.puts "\n======================================================================================================"
	out.puts "Random sampler to evaluate how many states and which cutoff time it needs to distinguish."
	out.puts "between parameter configurations."
	out.puts "Usage: Much like param_ils_1_0_run.rb: ruby random_sampler.rb, followed by the following optional arguments (defaults in [parentheses])"

	out.puts "General"
	out.puts "-------\n"
	out.puts "-algo <X>, where <X> calls (a wrapper of) the algorithm to be optimized. [#{@algo}]"
	out.puts "           This wrapper/algorithm has to follow the input/output format in README.TXT" 
	out.puts "-run_obj <X>, where <X> in {runtime, runlength, speedup} [#{@run_obj}] (or user-defined in file algo_specifics.rb)"
	out.puts "-overall_obj <X>, where <X> in {mean, adj_mean, median, q90, geomean} [#{@overall_obj}]\n"
	out.puts "-cutoff_time <X>, where <X> is a float [#{@cutoff_time}]"
	out.puts "-cutoff_length <X>, where <X> is an integer [#{@cutoff_length}]"
	out.puts "-N <X>, where X is an integer indicating the maximal number of samples obtained per config [#{@N}]."# (BasicILS(N) or FocusedILS with a bound on each N(theta)) [#{@N}]"
	out.puts "-numRun <X>, where X is an integer indicating the how manieth run of this tuning scenario this is[#{@numRun}]."
end

	
# =========================================
#                  MAIN
# =========================================

# =========================================
# defaults.
# =========================================
@run_obj = "runtime"
@overall_obj = "adj_mean"

@N = 100
@cutoff_time = 10
@cutoff_length = "max"

@param_file = "params.txt"
@numRun = 0

# =========================================
# Read in ParamILS-specific command line options.
# =========================================
0.step(ARGV.length-1, 2){|i|
	case ARGV[i]
	when "-algo"
		@algo = ARGV[i+1]
	when "-run_obj"
		@run_obj = ARGV[i+1]
	when "-overall_obj"
		@overall_obj = ARGV[i+1]

	when "-cutoff_time"
		@cutoff_time = ARGV[i+1]
	when "-cutoff_length"
		@cutoff_length = ARGV[i+1]

	when "-N"
		@N = ARGV[i+1].to_i
		
	when "-paramFile"
		@param_file = ARGV[i+1]

	when "-scenario"
		@scenario = ARGV[i+1]

		case @scenario
		when "SAT4J-SW"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = 10
			@cutoff_length = "max"

			@tunerTimeout = 36000
			@maxEvals = 20000

			@instance_seed_file = "sw-sat4j-5-to-10s-scrambled-first1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "sw-sat4j-5-to-10s-scrambled-last1000-det_algo/instance_list_0.txt" if @test

		when "SAT4J-SWorig"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = 5
			@cutoff_length = "max"

			@tunerTimeout = 36000
			@maxEvals = 20000

			@instance_seed_file = "sw-sat4j-1-to-2s-scrambled-first1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "sw-sat4j-1-to-2s-scrambled-last1000-det_algo/instance_list_0.txt" if @test

		when "SPEAR-swv"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "SWV-scrambled-first302-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat/instance_list_0.txt"

		when "SPEAR-bmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "BMC-scrambled-first64-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat/instance_list_0.txt"

		when "SPEAR-cbmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "BMC-scrambled-newsplit-train-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-newsplit-test-strat/instance_list_0.txt"

		when "SPEAR-obmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "BMC-scrambled-first64-minisattime-ordered/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat/instance_list_0.txt"

		when "SPEAR-ibm"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean1000"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 8640000
			@maxEvals = 200000

			@instance_seed_file = "IBM_40randomsets-scrambled-first382-feasible287-strat/instance_list_#{@numRun}.txt"
			#			@instance_seed_file = "IBM_40randomsets-scrambled-first382-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "IBM_40randomsets-scrambled-last383-strat/instance_list_0.txt"

		when "SAT4J-bmc"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean1000"
			
			@cutoff_time = 10 #300 
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "BMC-scrambled-first64-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat-det_algo/instance_list_0.txt"

		when "SAT4J-swv"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean1000"
			
			@cutoff_time = 10 #300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "SWV-scrambled-first302-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat-det_algo/instance_list_0.txt"

		when "SPEAR-pjava"
			@algo = "spear1.2.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 100000
			@maxEvals = 20000

			@instance_seed_file = "PJAVA-scrambled-first38-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "PJAVA-scrambled-last38-strat/instance_list_0.txt" if @test

		when "CPLEX-regions"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "adj_mean"

			@cutoff_time = 10
			@cutoff_length = "max"

			@tunerTimeout = 200000
			@maxEvals = 20000

			@instance_seed_file = "cats-regions-goods100-bids500-det_algo/instance_list_#{@numRun}.txt"
			#@instance_seed_file = "cats-regions-goods200-bids1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "XXX/instance_list_0.txt" if @test

		when "CPLEX-200regions"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "adj_mean"

			@cutoff_time = 10
			@cutoff_length = "max"

			@tunerTimeout = 200000
			@maxEvals = 20000

			@instance_seed_file = "cats-regions-goods200-bids1000-det_algo/instance_list_#{@numRun}.txt"
			#@instance_seed_file = "cats-regions-goods200-bids1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "XXX/instance_list_0.txt" if @test
			
		when "CPLEX-miplib"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean1000"

			@cutoff_time = 100
			@cutoff_length = "max"

			@tunerTimeout = 200000
			@maxEvals = 20000

			@instance_seed_file = "miplib3-scrambled-first33-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "miplib3-scrambled-last32-strat-det_algo/instance_list_0.txt" if @test
		
		when "SAPS-SW"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = "10"
			@cutoff_length = "max"
			
			@tunerTimeout = 36000
			@maxEvals = 20000

			@instance_seed_file = "SW-saps-5s-to-10s-scrambled-first113/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-5s-to-10s-scrambled-last100/instance_list_0.txt" if @test

		when "SAPS-SWorig"
			@algo = "saps"

			@run_obj = "runlength"
			@overall_obj = "median"

			@cutoff_time = "max"
			@cutoff_length = "2000000"
			
			@tunerTimeout = 36000
			@maxEvals = 20000

			@instance_seed_file = "SW-saps-1M-to2M-scrambled-first100/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-1M-to2M-scrambled-last82/instance_list_0.txt" if @test

		when "SAPS-SW100"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = "max"
			@cutoff_length = 2000000
			
			@tunerTimeout = 18000
			@maxEvals = 20000

			@instance_seed_file = "SW-saps-100s-to200s-scrambled-first231/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-100s-to200s-scrambled-last100/instance_list_0.txt" if @test

		when "SAPS-QWH"
			@algo = "saps"

			@run_obj = "runlength"
#			@overall_obj = "q90"
			@overall_obj = "median"

			@cutoff_time = 1
			@cutoff_length = "max"
			
			@tunerTimeout = 9000
			@maxEvals = 20000

			@instance_seed_file = "qwh_med_inst/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "qwh_med_inst_test/instance_list_0.txt" if @test

		when "SAPS-QWHorig"
			@algo = "saps"

			@run_obj = "runlength"
			@overall_obj = "q90"

			@cutoff_time = 1
			@cutoff_length = "max"
			
			@tunerTimeout = 9000
			@maxEvals = 20000

			@instance_seed_file = "qwh_med_inst/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "qwh_med_inst_test/instance_list_0.txt" if @test

		when "GLS4MPE"
			@algo = "gls4mpe"

			@run_obj = "prob"
			@overall_obj = "mean"

			@cutoff_time = 10
			@cutoff_length = "max"
			
			@tunerTimeout = 200000
			@maxEvals = 20000

			@instance_seed_file = "gridGH25GW25K2-pseudo-opt-scrambled-first50/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "gridGH25GW25K2-pseudo-opt-scrambled-last50/instance_list_0.txt" if @test

		when "SWGCP-all-Spear"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean10"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000
			@maxEvals = 20000

			@instance_seed_file = "SWGCP-all-scrambled-first1000-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWGCP-all-scrambled-last1000-strat/instance_list_0.txt" if @test

		when "SWGCP-all-Sat4J"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean10"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000
			@maxEvals = 20000

			@instance_seed_file = "SWGCP-all-scrambled-first1000-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWGCP-all-scrambled-last1000-strat-det_algo/instance_list_0.txt" if @test

		when "SWGCP-allSat-SAPS"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "mean10"

			@cutoff_time = 5
			@cutoff_length = "max"
			
			@tunerTimeout = 18000
			@maxEvals = 20000

			@instance_seed_file = "SWGCP-allSat-scrambled-first1000-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWGCP-allSat-scrambled-last1000-strat/instance_list_0.txt" if @test

		else
			raise "Undefined scenario #{@scenario}"
		end

	else
		output_help($stdout)
		puts "\n\n"
		raise "Unknown argument: #{ARGV[i]}"
	end
}

if ARGV.length < 1
	output_help($stdout)
	exit -1
end

@cutoff_time = 1000000000 if @cutoff_time == "max"
@cutoff_time = @cutoff_time.to_f

@cutoff_length = 2147483647 if @cutoff_length == "max"
@cutoff_length = @cutoff_length.to_i

#=== Read params from param file.
@params, @domain, @default, $conditionals, @forbidden_combos = read_params(@param_file) # Need conditionals as global to enable strip_state in param_reader
num_params = 0
for param in @params
	num_params += 1 if @domain[param].length>1
end

sample_filename = "1000samples-algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-time#{@cutoff_time}-length#{@cutoff_length}_#{@numRun}.txt"
@out = File.open(sample_filename, "w")

#=== Initialize variables.
seed = (@numRun+1)*1234 # Seed for Param_ILS, not the one to pass on to algorithms.
puts "seed: #{seed}"
#@out.puts "seed: #{seed}"
srand(seed)

1000.times{|i|
	state = init_random()
	output(state)
}
@out.close