require "param_reader.rb"
require "global_helper.rb" # for instance hash

# =========================================
#                  MAIN
# =========================================

srand 1234

#=== Deal with inputs.
if ARGV.length < 7 or ARGV.length > 8
	puts "READ_ALL runs all parameter settings for specified instances and runs."
	puts "Usage: ruby read_all.rb <algo> <within_obj> <numRuns> <cutoff_time> <cutoff_length> <instanceset_filename> <param_file> [<params to keep fix and their values, in quot marks>]."  
	puts "Examples:\nruby ../scripts/read_all.rb saps median_runtime 1 900 max sw-sat20.txt params.txt"
	exit -1
end
algo = ARGV[0]
within_obj = ARGV[1]
numRuns = ARGV[2].to_i
cutoff_time = ARGV[3]
cutoff_length = ARGV[4]
instanceset_filename = ARGV[5]
param_file = ARGV[6]
if ARGV.length > 7
	raise "still need to implement fixed parameters."
	#fix_input = ARGV[10]
end

out = File.open("readall-algo#{@algo}-#{within_obj}-runs#{@numRuns}-time#{$cutoff_time}-length#{$cutoff_length}_#{instanceset_filename}", "w")
string_arguments = ARGV.map{|x| "\"#{x}\""}.join(" ")
out.puts "Call: read_all.rb #{string_arguments}"

#=== Read params from the executable.
params, domain, default, $conditionals, forbidden_combos = read_params(param_file) 

sorted_params = params.sort{|x,y| $conditionals.key?(x) ? ($conditionals.key?(y) ? 0 : 1) : -1 }

#=== Build fixed parameters.
#set_fixed_params(fix_input)

all_distinct_states = get_all_distinct_states(sorted_params, domain, forbidden_combos, out)

###########################################################################
#=== Get instances.
###########################################################################
i_Hash, orig_instances_sorted = getInstanceHash(instanceset_filename, algo)
desired_quals = Hash.new
for inst in orig_instances_sorted
	desired_quals[inst] = i_Hash[inst]["desired_qual"]
end

numInstances = orig_instances_sorted.length
numberOfRunsUsedForAlgorunId = Hash.new

###########################################################################
#=== Read results. I only keep results for an instance if every parameter setting has been run.
###########################################################################
results = Hash.new
for instance_relname in orig_instances_sorted
	puts instance_relname
	results[instance_relname] = Hash.new
	
	complete = true
	for algo_params in all_distinct_states
		stripped_state, state_as_string = algo_params
		
		#=== Read results from DB.
		algorun_config_id, instance_id, algo_id = get_algorun_config_id(algo, stripped_state, instance_relname, desired_quals[instance_relname], cutoff_time, cutoff_length)
#		puts "#{algorun_config_id}, #{instance_id}, #{algo_id}"
		numOfResults = numberOfAlgoResults(algorun_config_id)
		if numOfResults < numRuns
			results.delete(instance_relname)
			puts "Skipping incomplete instance #{instance_relname}."
			complete = false
			break
		end
		thisResults = readAlgoResults(algorun_config_id, 1, numRuns)
		results[instance_relname][state_as_string] = []
		for res in thisResults
			results[instance_relname][state_as_string] << [res[1], singleInstanceObjective(algo, within_obj, [res], desired_quals[instance_relname]), instance_id, algo_id, ]
		end
	end
	next unless complete

	puts "Instance #{instance_relname} (#{instance_id}) complete."
	for algo_params in all_distinct_states
		stripped_state, state_as_string = algo_params
		
		sortedParams = stripped_state.keys.sort
		parameters = sortedParams.map{|param| stripped_state[param]}.join(", ")
		
		thisResults = results[instance_relname][state_as_string]
		for res in thisResults
			out.puts "#{res[0]}, #{res[1]}, #{res[2]}, #{res[3]}, #{parameters}"
		end
	end
end

#end
#getAlgoResultsForInstsAndParams(algo, instanceAndParamsHash, cutoff_time, cutoff_length, oncluster) if count > 0
#runAlgoCommandsOnCluster(algoCommands)
out.close