#require "param_ils_helper.rb"
require "global_helper.rb"
require "param_reader.rb"

    class Object
      def deep_clone
        Marshal::load(Marshal.dump(self))
      end
    end
    
###########################################################################
# Takes all results so far and outputs to the outfile: time, haluzinated_qual_mean, haluzinated_qual_std, real_qual_mean, real_qual_std
###########################################################################
def outputResults(results_of_all_runs, outfilename, outfilename_final, important_times)
	puts "Writing results to outfile #{outfilename_final}"
#	p results_of_all_runs
#	p "and times"
#	p important_times

	File.open(outfilename, "w"){|out| 
		for results in results_of_all_runs
			out.puts "New run"
			for run in results
				out.puts run.join(", ")
			end
		end
	}
	
	require "stats_ils.rb"
	times = important_times.sort
	times.uniq!
	puts "#{times.length} different time points."

	statsfile = File.open(outfilename_final + "-stats", "w")

#	p "#{times.length} times: #{times.join(" ")}"
	File.open(outfilename_final, "w"){|out|
		for time in times
#			p "time: #{time}"
			complete = true
			quals = []
			real_quals = []
			for runresults in results_of_all_runs
#				p "  results of a complete run:"
#				p runresults
#				raise "runresults is empty" if runresults.length == 0
				
				#=== Discard first term if runtime of second term is smaller than what we care about - i.e. keep only the part of the array from the last time that's smaller than the time we care about.
				while runresults.length > 1
					secondresult = runresults[1]
					if secondresult[0] <= time+1e-6
						runresults.shift
					else
						break
					end
				end
				
				#=== Prepare against bugs
				if runresults.length > 1
					secondresult = runresults[1]
					raise "Can't have two runresults left with smaller time than time #{time}, but do: #{runresults.join(" ")}" if secondresult[0] < time
				end
				
				#=== Gather info from the current result.
				firstresult = runresults[0]
				if firstresult[0] <= time+1e-6
					quals << firstresult[1]
					real_quals << firstresult[2]
				else
					#=== This run hasn't done anything by that time - don't include.
					quals << 10000000000
					real_quals << 10000000000
#					complete = false
#					break
				end
			end
#			next unless complete
#			out.puts "#{time}, #{quals.avg}, #{quals.std}, #{real_quals.avg}, #{real_quals.std}"
			statsfile.puts "train #{time}, #{quals.join(", ")}. #{quals.avg} +/- #{quals.std}"
			statsfile.puts "test  #{time}, #{real_quals.join(", ")}. #{real_quals.avg} +/- #{real_quals.std}"
			out.puts "#{time}, #{quals.median}, #{quals.quantile(0.1)}, #{quals.quantile(0.9)}, #{real_quals.median}, #{real_quals.quantile(0.1)}, #{real_quals.quantile(0.9)}"
			out.flush
#			out.puts "#{time}, [#{quals.join(",")}], [#{real_quals.join(",")}]"
		end
	}
	statsfile.flush
	statsfile.close
end

###########################################################################
#                                           MAIN
###########################################################################

if ARGV.length < 8
	puts "real_evaluate.rb: independently evaluates the parameter assignments in a file by checking additional runs\n"
	puts "Usage:\nruby real_evaluate.rb <infilename> <minseed> <maxseed> <instanceset_name> <param_file> <oncluster> <use_db> <deterministic_algo>"
	puts "Output will be to file <infilename>__min<minseed>_<max<maxseed>_set<instanceset_name>"
	exit -1
end

$user_frank = true
infilename = ARGV[0]
minseed = ARGV[1].to_i
maxseed = ARGV[2].to_i
instanceset_name = ARGV[3]
param_file = ARGV[4]
oncluster = ARGV[5].to_i
@db = ARGV[6]
$deterministic_algorithm = ARGV[7]
if ARGV.length > 8
	fix_input = ARGV[8]
end

#=== 0 is not necessary false !!
@db = false if @db == "0"
$deterministic_algorithm = false if $deterministic_algorithm == "0"

require "dbi_ils_accessor.rb" if @db

if infilename =~ /-algo([^-]*)-/
	algo = $1
else raise "need -algo<algo> in infilename #{infilename}"
end

if infilename =~ /-within_obj([^-]*)-/ or infilename =~ /-runobj([^-]*)-/
	within_obj = $1
else raise "need -within_obj<within_obj> in infilename"
end

if infilename =~ /-obj([^-]*)-/ or infilename =~ /-overallobj([^-]*)-/
	obj = $1
else raise "need -obj<obj> in infilename"
end

cutoff_time = 10
cutoff_length = "_"

if infilename =~ /time(#{float_regexp})/
	cutoff_time = $1.to_f
end
if infilename =~ /length(#{float_regexp})/
	cutoff_length = $1
end
if infilename =~ /len(#{float_regexp})/
	cutoff_length = $1
end

cutoff_length = "2147483647" if cutoff_length == "_"
cutoff_length = "2147483647" if cutoff_length == "max"
cutoff_length = cutoff_length.to_f

infile = File.open(infilename)
system "mkdir real"
outfilename_final = "real/joint_med_#{infilename.sub(/\.txt/, "")}__min#{minseed}_max#{maxseed}_set#{instanceset_name}.txt".sub(/ /,"")
outfilename = "real/real_#{infilename.sub(/\.txt/, "")}__min#{minseed}_max#{maxseed}_set#{instanceset_name}.txt".sub(/ /,"")

outfilename_results = outfilename + "-results"
result_file = File.open(outfilename_results, "w")
result_file.close
result_file = File.open(outfilename_results, "a")


$total_cputime = 0

#@algo = "\"#{algo}\""
@algo = algo
@params, @domain, @default, $conditionals, @forbidden_combos = read_params(param_file) # Need conditionals as global to enable strip_state in param_reader

results_of_all_runs = []
runresults = nil
important_times = []
first = true
realQual = nil
runresults = []
saved_results = Hash.new
someResultsReady = false
while line = infile.gets
	if line =~ /Run/
		unless first 
			results_of_all_runs << runresults 
#			outputResults(results_of_all_runs.deep_clone, outfilename, outfilename_final, important_times) unless oncluster
		end
		runresults = []
		first = false
	else
		line = line.split(",").map{|x|x.strip}
		cpu_time = line[0]
		qual = line[1]
#		real = line[2]
		numEvaluations = line[2]
		numIterations = line[3]
		numFlips = line[4]

		params = line[5...line.length]
#puts line
		p params
		puts "Evaluating parameter setting #{params.join(", ")}" 
		if params.length <= 1
			tmp = params[0]
			params = {"param_string"=>tmp}
			state_as_string = tmp.dup
		else	
			param_array = params.dup
			
			params = Hash.new
			for i in 0...@params.length
				params[@params[i]] = param_array[i];
			end

			params = strip_state(params)
			state_as_string = state_string(params)
		end

		#=== Use the same evaluation runs for evaluating the different runs of the algorithm - get rid of this if clause to evaluate each one by itself.
		if saved_results.key?(state_as_string)
			realQual = saved_results[state_as_string]
		else
			puts "get results for instances file"
			instances, someResultsReady = getResultsForInstanceFile(@algo, params, cutoff_time, cutoff_length, minseed, maxseed, instanceset_name, state_as_string, oncluster, @db)
			puts "after get results for instances file"

			if someResultsReady
				for entry in instances
					res = entry["resultForState"][state_as_string]
					p res
					result_file.puts "#{entry["name"]} #{res[0]} #{res[1]} #{res[2]} #{res[3]} #{res[4]} #{res[5]}"
					result_file.flush
				end

				require "algo_specifics.rb"
#				realQual, numInstances = getGlobalObjective(@algo, within_obj, obj, instances, state_as_string, 300, cutoff_length)
				realQual, numInstances = getGlobalObjective(@algo, within_obj, obj, instances, state_as_string, cutoff_time, cutoff_length)
				saved_results[state_as_string] = Hash.new unless saved_results.key?(state_as_string)
				saved_results[state_as_string][cutoff_time] = realQual
			else
				saved_results[state_as_string] = -1
			end
		end

		important_times << cpu_time.to_f
		runresults << [cpu_time.to_f, qual.to_f, realQual.to_f]
		p [cpu_time.to_f, qual.to_f, realQual.to_f]
	end
end

result_file.close
results_of_all_runs << runresults # last run
p outfilename
outputResults(results_of_all_runs, outfilename, outfilename_final, important_times) if someResultsReady