require "English"
require "stats.rb"
require "cmd_line.rb"
require "hash.rb"
require "helper.rb"
require "jobber.rb"
require "exp.rb"

def read_params() #==== Reads parameters from executable.
	$param_values = Hash.new
	$conditionals = Hash.new
	$default = Hash.new
	IO.popen("#{sls_executable} -params"){|file|
		while line = file.gets and line =~ /^-(.*)\s*\{(.*)\}\s*\[(.*)\]\s*$/
			puts line
			parname, vals, default = $1, $2, $3
			values = vals.strip.split(",").map{|x|x.strip}
			unless values.include?(default)
				if default =~ /(#{float_regexp})/
					def_f = $1.to_f
					for val in values
						default = $1 if val =~ /(#{float_regexp})/ and $1.to_f == def_f
					end
				end
			end
			puts "default value #{default} is not a possible value for #{parname}. Only #{values.join(",")} are ok." unless values.include?(default)
	
			$param_values[parname] = values.map{|x| x.strip.to_s}
			$default[parname] = default
		end
		p $param_values
		$params = $param_values.keys.sort
	
		#already read line "Conditionals"
		# example: parameter a is only relevant if c=1: "a|c=1"
		while line=file.gets
			line =~ /^(.*)\|(.*) in \{(.*)\}/
			conditional_param, deciding_param, deciding_vals = $1, $2, $3
			conditional_param.strip!
			deciding_param.strip!
			deciding_vals = deciding_vals.strip.split(",").map{|x|x.strip}
			
			unless $param_values.key?(conditional_param) 
				puts "Error. Conditional parameter #{conditional_param} does not exist." 
				next
			end
			unless $param_values.key?(deciding_param)
				puts "Error. Variable #{deciding_param} which #{conditional_param} is conditional on does not exist." 
				next
			end
			deciding_vals.each{|x| 
				unless $param_values[deciding_param].include?(x)
					puts "Error. Value #{x} of variable #{deciding_param} which #{conditional_param} is conditional on does not exist."
					next
				end
			}
#			deciding_vals.map!{|x| $param_values[deciding_param].index(x)}
			$conditionals[conditional_param] = Hash.new unless $conditionals.key?(conditional_param)
			$conditionals[conditional_param][deciding_param] = deciding_vals
		end
	}
end


def build_start(start)
	@start_values = Hash.new
	if start
		start = start.split.map! {|x| x.strip} 
		0.step(start.length-1,2){|i|
			param = start[i]
			value = start[i+1]
			@start_values[param]=value
		}
	end
	output "Manual start values: #{@start_values}"
end

def build_fixed_ass(fix, opt, param2optimize) #====== Gather parameters that are fixed and default params that are not to be optimized.
	@fixed_ass = Hash.new
	#======= Parse fix and remember them in hash.
	if fix 
	   p fix
		fix = fix.split.map! {|x| x.strip.to_s} 
		0.step(fix.length-1,2){|i|
			param = fix[i]
			value = fix[i+1]
			p $param_values
			p $param_values[param]
			p fix
			p i
			unless $param_values[param].include?(value)
				puts "param #{param} can only have values #{$param_values[param].to_a.join(", ")}, not #{value}"
				exit -1
			end
			@fixed_ass[param] = value
		}
	end
	output "manually fixed: #{fixed_string}"
	
	#======= For non-mentioned params assign their default value as fix.
	for param in $params
		if not opt.include?(param) and not @fixed_ass.key?(param) and not param2optimize == param
			@fixed_ass[param] = $default[param]
		end	
	end
	output "with default values fixed: #{fixed_string}"
end


def state_string(state)
	$params.map{|param| param + "=" + state[param]}.join(" ")
end

def output_result(result)
#	for test_inst in @test_instance_names
	for test_inst in @test_instances
		output get_instance_name(test_inst) + ": " + result[test_inst].join(" ") + ", opt: #{get_optimal_mpe(get_instance_name(test_inst), 0)}"
	end
	output "Overall goodness: #{goodness(result).join(", ")}"
end

def strip_state(state)
	result = Hash.new
	for param in state.keys
		if $conditionals.key?(param)
			preconditions_ok = true
			for key in $conditionals[param].keys
				preconditions_ok = false unless $conditionals[param][key].index(state[key])
			end
			next unless preconditions_ok
		end
		result[param] = state[param]
	end
	return result
end

def result_as_string(result)
	quals = []
#	for test_inst in @test_instance_names
	for test_inst in @test_instances
		opt = get_optimal_mpe(get_instance_name(test_inst), 0)
		qual = result[test_inst][0]
		quals << "%.5f [%.5f]" % [qual, opt==0 ? -1 : qual.to_f/opt.to_f]
	end
	return quals.join(" ") + " --> [#{goodness(result).join(", ")}]"
end

def goodness(result)
	#returns [avg. relative qual. over best, percent solved]
	n=0
	num_found = 0

	overall = 0
	for test_inst in @test_instances 
          #puts "test inst #{test_inst}"
          avg = result[test_inst][1]
		opt = get_optimal_mpe(get_instance_name(test_inst), 0)
		overall += 10**(avg-opt)
          #puts "opt = #{opt}, avg=#{avg}"
          n+=1
		num_found += 100.0*result[test_inst][2]
	end
#p overall
#p num_found
  return [overall *100.0 / n, num_found/n] 

=begin
	overall = 0
	for test_inst in @test_instances
		avg = result[test_inst][1].to_f
		opt = get_optimal_mpe(get_instance_name(test_inst), 0)
		overall += (opt/avg)
		n+=1
		num_found += 100.0*result[test_inst][2]
	end
	return [overall *100.0 / n, num_found/n] 
=end
=begin
	overall = 1
	for test_inst in @test_instances
		avg = result[test_inst][1].to_f
		opt = get_optimal_mpe(get_instance_name(test_inst), 0)
		if opt == 0
			overall *= (-avg) # high negative averages are bad, small negative avgs are better.
		else
			overall *= (avg/opt)
		end
		
		n+=1
		num_found += 100.0*result[test_inst][2]
	end
	return [100*(overall **(1.0/n)-1) , num_found/n] 
=end	
end

def sls_param_string(param_state, t)
	state = strip_state(param_state)
	local_params = state.keys.sort
	param_string = " " + local_params.map{|param| "-#{param} #{state[param]}"}.join(" ")
	param_string << " -x #{@numruns} -t #{t}"
	return param_string
end

def eval(param_state,t)
	param_string = sls_param_string(param_state, t)
	state = strip_state(param_state)
	return @cached[state][t] if @cached.key?(state) and @cached[state].key?(t)
	result = Hash.new
	data = Hash.new
	for test_inst in @test_instances
		directory, inst_name = File.split(test_inst)
		inst_name.sub!(/\.simple/,"")
		dir = directory + "/lsf#{@lsf_model}"
		Dir.mkdir(dir) unless File.exist?(dir)
		resname = dir + "/#{inst_name}___sls#{param_string.gsub(/ /,"")}.res"
		outname = resname.sub(/\.res$/,".out")
		using_cached = false
		if File.file?(resname)
			File.open(resname){|file|
				content = file.gets(nil)
				using_cached = true if content =~ /Total time (#{float_regexp})/
			}
#			puts "Don't need to perform the SLS. #{resname} already exists."
		end

		opt = get_optimal_mpe(get_instance_name(test_inst), 0)
		data[test_inst] = [using_cached, opt, outname]
	end
	
	for test_inst in @test_instances
		using_cached, opt, outname = data[test_inst]
		unless using_cached
			#command = sls_executable + " #{param_string} -i #{test_inst} -o #{outname} #{opt ? "--optimalLogMPE " + opt.to_s : ""} "
			command = sls_executable + " #{param_string} -i #{test_inst} #{opt ? "--optimalLogMPE " + opt.to_s : ""}"
			#@jobQueue << getJob(command, outname, opt, true) # with special output handler
			@jobQueue << getJob(command, outname, opt)
		end
	end
	run_jobs(@jobQueue, @jobServerStatsfile) unless @jobQueue.empty?
	
	for test_inst in @test_instances
		using_cached, opt, outname = data[test_inst]
		resname = outname.sub(/\.out$/,".res")
		unless using_cached #File.file?(resname)
			File.open(outname){|outfile|
				File.open(resname, "w"){|resfile|
					content = outfile.gets(nil)
					content =~ /Worst solution quality found: (#{float_regexp})/
					raise "no worst solution quality listed in file #{outname}" if $1 == nil
					resfile.puts "Worst solution quality " + $1
					content =~ /Average solution quality found: (#{float_regexp})/
					resfile.puts "Average solution quality " + $1
					content =~ /(\d+)\/(\d+) runs found optimal MPE./
					resfile.puts "Number Solved/Number of runs " + $1 + "/" + $2
					content =~ /Experiment took (#{float_regexp}) seconds/
					resfile.puts "Total time " + $1
				}
			}
			File.delete(outname) if File.file?(outname)
		end

		result[test_inst] = []
		File.open(resname){|resfile|
			content = resfile.gets(nil)
			content =~ /Worst solution quality (#{float_regexp})/
			result[test_inst] << $1.to_f
			content =~ /Average solution quality (#{float_regexp})/
			result[test_inst] << $1.to_f
			content =~ /Number Solved\/Number of runs (\d+)\/(\d+)/
			result[test_inst] << ($1.to_f/$2.to_f)
			if ($1.to_f/$2.to_f).nan?
				p $1
				p $2
				p resname
				p content
				raise "solved ratio is nan" 
			end
			content =~ /Total time (#{float_regexp})/
			result[test_inst] << $1.to_f
		}
			
	end
	@cached[state] = Hash.new unless @cached.key?(state)
	@cached[state][t] = result
	return result # worst, avg, solved ratio, total time
end

=begin
def cmp_single_results(res1,res2) 
#	return res1[0] / res2[0] # small: first is better ! (e.g. -1 / -5 = 0.2)

# -1: first better, 1: second better, 0: equal.
	return -1 if res1[2] > res2[2] # number of sol. found
	return 1 if res1[2] < res2[2]

	return -1 if res1[1] > res2[1] # avg qual (want high avg)
	return 1 if res1[1] < res2[1]

	return -1 if res1[2] < res2[2] # total  time (want low time)
	return 1 if res1[2] > res2[2]

	return -1 if res1[0] > res2[0] # min qual (want high min qual)
	return 1 if res1[0] < res2[0]

	return 0 
end
=end

def diff_number_runs_solved(res1,res2)
# negative if first solved more.
	return res2[2] - res1[2]
end

def cmp_results(result1, result2) # <1: first better
	#score = 1
	good1 = goodness(result1)
	good2 = goodness(result2)
	solved_counts = 0.5
#solved_counts = 1
		
	if good1[0] == 0
		if good2[0] == 0 
			qual_ratio = 1
		else
			qual_ratio = 1e100
		end
	else 
		qual_ratio = (good2[0]/good1[0])
	end

	
	if good1[1] == 0
		if good2[1] == 0 
			solved_ratio = 1
		else
			solved_ratio = 1e100
		end
	else 
		solved_ratio = (good2[1]/good1[1])
	end
	
	score = (1-solved_counts)* qual_ratio  + (solved_counts) * solved_ratio # < 1: first is better. 

=begin	score = 0
	for test_inst in @test_instances
		cmp = cmp_single_results(result1[test_inst], result2[test_inst])
#		puts "cmp( #{result1[test_inst][0]} , #{result2[test_inst][0]} ) = #{cmp}"
#		score *= cmp
		score += cmp
	end
	if score == 0 
		for test_inst in @test_instances
			cmp = diff_number_runs_solved(result1[test_inst], result2[test_inst])
			score += cmp
		end
		score = -0.5 if score < 0
		score = 0.5 if score > 0
	end
=end
	return score
end

def poss_flips(current_state)
	p "state:"
	p current_state
	state = strip_state(current_state)
	p "stripped state:"
	p state
	result = []
	for param in state.keys
		for value in $param_values[param]
			result << [param, value] unless current_state[param] == value
		end
	end
	return result
end

def output(s)
	puts s
	@out.puts s
	@out.flush
end

def iterated_local_search(minT, maxT)
=begin
	ts = []
	t = 0.125
	while t < maxT
		ts << t if t >= minT
		t *= 2
	end
	ts << maxT
#	for t in ts # iterative increasing of time bound s.t. only few computations remain with large time bound.
=end
	t = maxT

	current_state = init()
  iteration = 1
	last_ils_state = basic_local_search(current_state, t, iteration)
	current_state = last_ils_state.dup
  
  last_improving_iteration = iteration
  last_imp_flip = @numflip-1
  iteration += 1
	while iteration <= @number_of_params_to_opt*3
          current_state = pertubation(current_state, 3)
          current_state = basic_local_search(current_state, t, iteration)
		
          acc_state =  acceptance_criterion(last_ils_state, current_state,t)
          if @accepted_last == "acc"
            last_improving_iteration = iteration
            last_imp_flip = @numflip
          end

          saveResultOfIteration(current_state,t,iteration,true)
          current_state = acc_state
          last_ils_state = current_state

          iteration += 1
	end
	return [last_ils_state, eval(last_ils_state,t), last_imp_flip]
end

def init()
	init_state = Hash.new
	for param in $params
		init_state[param] = $default[param]#rand(@domsize[param])
	end
	for key in @fixed_ass.keys
		init_state[key] = @fixed_ass[key]
	end
	for key in @start_values.keys
		init_state[key] = @start_values[key]
	end
	return init_state
end

def pertubation(state, strength)
	result_state = state.dup
	strength.times{|x|
		possible_flips = poss_flips(result_state)
		rand_flip_num = rand(possible_flips.length)

		param,val = possible_flips[rand_flip_num]
		redo if @fixed_ass.key?(param)
		output "perturbing param #{param} to val #{val}"
		result_state[param] = val
	}
	return result_state
end

def acceptance_criterion(last_ils_state, current_state, t)
  if state_string(last_ils_state) == state_string(current_state)
		output "same state as last ILS: #{state_string(last_ils_state)}"
		@accepted_last = "same"
		return current_state
	end

#	output "score cmp(last, curr) = #{cmp_results(eval(last_ils_state,t), eval(current_state,t))}"
#	if cmp_results(eval(last_ils_state,t), eval(current_state,t)) == 1
#		output "accepting equally good state: #{state_string(current_state)}"
#		@accepted_last = "equal"
#		return current_state
#	end
	
	if cmp_results(eval(last_ils_state,t), eval(current_state,t)) > 1
		output "accepting better #{state_string(current_state)}"
		@accepted_last = "acc"
		return current_state
	end

	output "rejecting worse #{state_string(current_state)}, going back to #{state_string(last_ils_state)}"
	@accepted_last = "rej"
	return last_ils_state
end

#==== Local search
def basic_local_search(start_state, t, iteration)
	current_state = start_state

	saveResultOfIteration(current_state,t,iteration)

	p current_state
	output "   BLS in iteration #{iteration}, start with " + state_string(current_state) + "\n   Algo : " + sls_param_string(current_state, t) + "\n   with result " + result_as_string(eval(current_state, t)) + "\n\n"
	visited = [current_state.clone]
	while true # local search.
          old_result = eval(current_state, t)
		
		best_score = 1 # remains the same for additive scores
		best_flips = []
		possible_flips = poss_flips(current_state)
		for poss_flip in possible_flips
			param = poss_flip[0]
			i = poss_flip[1]
			next if @fixed_ass.key?(param)
			old_val = current_state[param]
			output("        " + param + "->#{i}       (#{state_string(current_state)})")
			
			current_state[param] = i
			if visited.include?(current_state) # disallow cycling.
				output("          -> not allowed, already visited.\n\n")
				current_state[param] = old_val
				next
			end
			result = eval(current_state, t)
			
			score = cmp_results(result, old_result) # best improvement
			output "          -> result: #{result_as_string(result)}, #{score}\n\n"

			best_flips << [param,i] if score < best_score + @@eps
				
			# only best move so far
			if score < best_score - @@eps
				best_score = score
				output "   found new best score #{best_score} with #{param}=#{i} (#{state_string(current_state)})\n\n"
				best_flips = [[param,i]]
			end
			current_state[param] = old_val
		end

=begin
		# TODO: put a "if was local minimum" here
			possible_flips2 = poss_flips(current_state)
			for poss_flip2 in possible_flips2
				param2 = poss_flip2[0]
				j = poss_flip2[1]
				next if @fixed_ass.key?(param2)
				old_val2 = current_state[param2]
				current_state[param2] = j
						
				if visited.include?(current_state) # disallow cycling.
#						puts " not allowed, already visited."
					next
				end
				result = eval(current_state, t)
				score = cmp_results(result, old_result) # best improvement
				@out.puts "      " + param2 + "=#{j}: #{result_as_string(result)}, #{score}"
				@out.flush
				puts "      " + param2 + "=#{j}: #{result_as_string(result)}, #{score}"
				# another best move.
				best_flips << [param,i, param2,j] if score < best_score + @@eps
					
				# only best move so far
				if score < best_score - @@eps
					best_score = score
					best_flips = [[param,i, param2,j]]
				end
				current_state[param2] = old_val2
			end
			current_state[param] = old_val
		end
=end
		# Actually flip variable.
		output "     Evaluated neighbourhood. best_score = #{best_score}"
		if best_score <= 1-@@eps
			randbest = rand(best_flips.length)
#				param,val, param2, val2 = best_flips[randbest]
			param,val = best_flips[randbest]

			current_state[param]  = val
			saveResultOfIteration(current_state,t,iteration)
#				current_state[param2]  = val2
			output "\n              ---> Flipping #{param} to #{val}"
			output "              new state:      " + state_string(current_state) + " with result " + result_as_string(eval(current_state,t)) +"\n\n"
		else 
			break
		end
		visited << current_state.clone
	end
	output "   LM for iteration #{iteration} at time #{t}: " + state_string(current_state) + "\n with result " + result_as_string(eval(current_state, t))
	output "\n========== DETAILED RESULTS (iteration #{iteration}): =========="
	output_result(eval(current_state, t))
	output "================================================\n"
	return current_state
end

def fixed_string
	fixed_string = ""
	for param in @fixed_ass.keys
		fixed_string << "#{param}=#{@fixed_ass[param]} "
	end
	return fixed_string
end

def saveResultOfIteration(state,t, iteration,*print_acceptance)
	if print_acceptance and print_acceptance[0]
		@numflip -= 1
		iteration = "#{iteration} - #{@accepted_last}"
	end

	stripped_state = strip_state(state)
	@ls_res[@numflip] = Hash.new

	@ls_res[@numflip]["Parameters"] = Hash.new
	local_params = stripped_state.keys
	for param in local_params
		@ls_res[@numflip]["Parameters"][param] = state[param] unless @fixed_ass[param]
	end

#	result = eval(state, t)
#	avg_qual = 0
#	for test_inst in @test_instances 
#		avg = result[test_inst][1]
#		opt = get_optimal_mpe(get_instance_name(test_inst), 0)
#		approx_qual = 10**(avg-opt)
#		avg_qual += approx_qual
#	end
#	avg_qual /= @test_instances.length
	
	good = goodness(eval(state,t))
#	@ls_res[@numflip]["t"] = t
	@ls_res[@numflip]["Solved"] = good[1]
	@ls_res[@numflip]["Avg. Quality"] =  good[0]#.to_s + "\\%"
	@ls_res[@numflip]["Iteration"] = iteration

	write_table(@ls_res_filename, @ls_res)
	@numflip += 1
  #			call "ruby -S hash2tex.rb -i #{@ls_res_filename} -all 
end

# #########################
# =========================================
#                         MAIN
# =========================================
# #########################

#=======  Create outfile
t = Time.now
@lsf_model = 3
@lsf_model = read_int("-lsf") if check_arg("-lsf")
now= t.strftime("%Y-%m-%dat%I-%M%p")
#@out = File.open(path+"tune-#{now}.txt", "w")
@out = File.open(path+"tune-#{@lsf_model}.txt", "w")
@ls_res_filename = path + "results/ls-tuner-#{@lsf_model}.hash"

arguments = []

cmd_line_call = File.basename($0).to_s + " " + $ARGV.join(" ")
output "CALL: " + cmd_line_call

#======= Jobserver stuff.
@jobServerStatsfile = "jobserver_stats-tuner-#{@lsf_model}.txt"
jobserver_config(@lsf_model)
@jobQueue = []

#=======  Read input.
@test_instances = read_in('\.simple$')
directory = File.dirname(@test_instances[0])
output "instances to optimize on: " + @test_instances.join(", ")
#@test_instance_names = @test_instances.map{|x|File.basename(x).sub(/\.simple/,"")}
unless @test_instances.length > 0
	puts "No test instances #.\nExiting."
	exit -1
end
minT, maxT = 0.2, 30.0
minT = read_float("-mint") if check_arg("-mint")
maxT = read_float("-maxt") if check_arg("-maxt")
@numruns = 5
@numruns = read_int("-x") if check_arg("-x")

#======= Read params from the executable.
read_params()
output "conditionals begin: "
for key in $conditionals.keys.sort
	output "   #{key}| #{$conditionals[key].to_a.join(" ")}"
end
output "conditionals end. "

#======= Read in fixed params and those to optimize.
opt = read_all_after("-opt")
output "-opt " + opt.join(" ").to_s
@number_of_params_to_opt = opt.length

fix = read_smth("-fix")
output "-fix \"#{fix}\""

#======= Gather fixed assignments (also defaults)
param2optimize = read_smth("-paramopt") if check_arg("-paramopt")
build_fixed_ass(fix, opt, param2optimize)

#======= Gather manual start values
start = read_smth("-start")
build_start(start)

#=======  Init current_state.
@@eps = 1e-6
current_state = Hash.new
srand 1234
best_params = []
best_results = []
@cached = Hash.new

@ls_res = Hash.new
@numflip = 1

#if param2optimize
#  algos = []
#  p $param_values[param2optimize]
#  for val in $param_values[param2optimize]
#    output "\n Optimizing for #{param2optimize}=#{val} with fixed params " + fixed_string
#    @fixed_ass[param2optimize] = val
#    best_state, best_result = iterated_local_search(minT, maxT)
#    algos << "sls" + sls_param_string(best_state, maxT)
##call "ruby -S hash2tex.rb -i #{@ls_res_filename} -all -rowtitle Iteration -noland &"
#  end
#  output "===================================================================="
#  output "     PARAMS FOR ALL VALUES OF #{param2optimize} OPTIMIZED"
#  output "===================================================================="
#  algos_string = algos.map{|x|"\"" + x + "\""}.join(" ")
#  experiment_command = "ruby -S exp.rb -i* #{directory} -name #{param2optimize}-#{File.basename(directory)}-tune -algos #{algos_string} -lsf #{@lsf_model}"
#  output "Calling experiment as follows in the next line:\n" + experiment_command 
#  call experiment_command, 2
#

output "\n Single run with fixed parameters: " + fixed_string
best_state, best_result, last_improving_flip = iterated_local_search(minT, maxT)
output "================================"
output "     BEST ASSIGNMENT FOUND"
output "================================"
output state_string(best_state)
output "It yields the following results at time #{maxT}:"
output_result(best_result)
output "======================================"
output "Search for parameter setting finished."
output "======================================"
  
=begin	#===== For every parameter that can be flipped from best_state, do an experiment showing how all other local choices are worse.
	algo = "sls" + sls_param_string(best_state, maxT)
	state = strip_state(best_state)
	for param in state.keys
		next if @fixed_ass.key?(param)
		next if $param_values[param].size == 1
		algos = []
		poss_params = $param_values[param]
		poss_params << state[param] unless poss_params.include?(param) # for defaults that are not eligible input.
		for newval in poss_params
			old_val = state[param]
			best_state[param] = newval
			best_state[param] = old_val
		end
		# calling the experiment
		part_algo = algo.sub(/-#{param} #{float_regexp}/, "")
		experiment_command = "ruby -S exp.rb -i* #{directory} -name #{param}-#{File.basename(directory)}-tuner -cmd \"#{part_algo}\" -param #{param} -lsf #{@lsf_model} -cached"
		output "Calling experiment with command in the following line:\n" + experiment_command
		call experiment_command, 2
#		do_experiment(directory, "#{param}-#{File.basename(directory)}-tuner", param, algos, @test_instances, true, false, false, 2, false)
	end
=end

@out.close

#=== Highlight row in which optimum was found first.
table = read_table(@ls_res_filename)
for key in table[last_improving_flip].keys
  table[last_improving_flip][key] = "\\textbf{#{table[last_improving_flip][key]}}" unless table[last_improving_flip][key].kind_of? Hash
end
write_table(@ls_res_filename, table)

call "ruby -S hash2tex.rb -i #{@ls_res_filename} -all -rowtitle Flip -noland -caption \"Results of ParamILS for algorithm #{fixed_string} on problem sets \\bn{bnrep} and \\bn{rand} with runtime \\$t=#{maxT}\\$ seconds.\" -label \"ILSParam-#{fixed_string}\""

