require "dbi_ils_accessor.rb"
   class Object
      def deep_clone
        Marshal::load(Marshal.dump(self))
      end
    end
    
if ARGV.length == 2 and ARGV[1] == "-params"
	puts "alpha {1.01, 1.066, 1.126, 1.189, 1.256, 1.326, 1.4} [1.189]" #uniform on log scale between 1.01 and 1.4
	puts "rho {0, 0.17, 0.333, 0.5, 0.666, 0.83, 1} [0.5]" #uniform
	puts "ps {0, 0.033, 0.066, 0.1, 0.133, 0.166, 0.2} [0.1]" #uniform
	puts "wp {0, 0.01, 0.02, 0.03, 0.04, 0.05, 0.06} [0.03]" #uniform
	exit
end

qual = 0
time = 10
length = "max"
runs = 10
#=== Deal with inputs.
if ARGV.length < 4
	puts "saps_wrapper.rb is a wrapper for the SAPS algorithm."
	puts "Usage: ruby saps_wrapper.rb <instance_relname> <run_objective> [optional parameters] -params \"<params>\"."
	puts "Optional extra parameters (in any order, followed by the value for that parameter, default in parentheses):"
	puts "-qual <qual>(#{qual}) -time <time>(#{time}) -length <length>(#{length}) -runs <runs>(#{runs})"
	puts "<params> has the form -param_name1 <param_value1> -param_name2 <param_value2> ..."
	puts "Output: list of objectives achieved in each run (e.g. runtime, runlength)"
	puts "Example: ruby saps_wrapper.rb instances/satlib/I-Hoos96b/flat175-417/flat175-7.cnf runlength -time 1 -runs 100 -alpha 1.189 -rho 0.5 -ps 0.133 -wp 0.04"
	exit -1
end

instance_relname = ARGV[0]
@within_obj = ARGV[1]
	#=== No default parameters needed - it will be assured that the algorithm call assigns every parameter. (e.g. to its default)
	
	paramkeys = ["alpha", "rho", "ps", "wp"] # TODO: Fix redundancy. This could be automatically detected from runnning this script with -params option.
	
	#=== Parse inputs
	for idx in 0...ARGV.length
		if ARGV[idx] ~= /$-(.*)/
			rest = $1
			if paramkeys.include?(rest)
				params[rest] = ARGV[idx+1] 
			end
		end
		case ARGV[idx] 
			when "-time" then time = ARGV[idx+1].to_f
			when "-length" then length = ARGV[idx+1]
			when "-qual" then qual = ARGV[idx+1].to_f
			when "-runs" then runs = ARGV[idx+1].to_i
		end
	end

	db = false
	#=== Create algorithm command
	algorun_cmd, cutoff_length,algo_config_id = buildAlgorunCmd(algo, params, instance_relname, qual, time, length, db)
	
	if db
		#=== If result(s) are in DB, output them and exit. BIG TODO: make sure the counters are increased and I don't get the same runs all the time!!
	else
		paramstring = paramkeys.map{|x| "-#{x} #{params[x]}"}.join(" ")
		cmd = "./ubcsat -alg saps #{paramstring} -inst #{instance_relname} -cutoff #{length} -timeout #{time} -r out stdout run,found,seed,best,beststep,steps,time"
		
		
		#=== Extract result
	
	#=== Output result(s)
	
end
	
	
    
###########################################################################
#Build algorun_cmd
###########################################################################
def run_saps
	
	buildAlgorunCmd(algo, params, instance_relname, desired_solqual, cutoff_time, cutoff_length, db=true)
	algo_config_id = 0

	raise "when getting to buildAlgorunCmd, params has to be a Hash. It is #{params}" unless params.is_a?(Hash)
	if algo == "saps"
		#=== Define default parameters that are used unless specifed differently.
		defaults = {"alpha"=>1.3, "rho"=>0.8, "ps"=>0.05, "wp"=>0.01}
		 = ""
		
		cutoff_length = 2147483647 if cutoff_length == "max"  #Want to save an integer in the DB
	elsif algo == "branin_function"
		#=== Define default parameters that are used unless specifed differently.
		defaults = {"x"=>2.5, "y"=>7.5}
		executable = "ruby exponential_function_with_branin_fucntion_as_median.rb"
		cmd_suffix = "#{cutoff_time}"
	elsif algo == "sls4mpe"
		#=== Define default parameters that are used unless specifed differently.
		defaults = {} # I think this is redundant !
		executable = "./sls4mpe"
		cmd_suffix = "-i #{instance_relname} --optimalLogMPE #{desired_solqual} -t #{cutoff_time}"
	else	
		raise "Need to specify executable and default parameters for algo #{algo} in function buildAlgorunCmd of algo_specifics.rb"
	end

=begin
#=== Generic treatment for each algo: check parameters and fill unspecified parameters with their default settings.
	for key in params.keys 
		raise "Unknown parameter #{key} for #{algo}" unless defaults.key?(key)
	end
	for key in defaults.keys
		params[key] = defaults[key] unless params.key?(key)
	end
=end

	param_string = params.keys.sort.map{|key| "-#{key} #{params[key]}"}.join(" ")
	algorun_cmd = "#{executable} #{param_string} #{cmd_suffix}"

	if db
		if algo == "saps"
			algo_config_id = get_saps_config_id([params["alpha"], params["rho"], params["ps"], params["wp"]])
		else
			algo_config_id = get_general_algorithm_config_id(algo, param_string)
		#	raise "Need to implement method to get algo_config_id for algo #{algo} in function buildAlgorunCmd of global_helper.rb"
		end
	end

	#=== Build command to run. (concat pairs of "-param param_value")
	return [algorun_cmd, cutoff_length, algo_config_id]
end


###########################################################################
# Run algorithm and save its result to DB.
###########################################################################
def runalgo(algorun_config_id, numRuns)
	algo, algorun_cmd = get_algo_and_runcmd_for_algorunconfig(algorun_config_id)
#	File.popen("pwd"){|file| puts "Dir = #{file.gets.chomp}"}
	
	t = Time.now
	datetime = t.strftime("%Y-%m-%d %H:%M:%S") # YYYY-MM-DD HH:MM:SS
	algo_output_file = "#{tmpdir}/tmp-#{datetime}-#{rand}".gsub(/ /,"")
    
	if algo == "saps"
		#=== SAPS -> Call special code to parse non-trivial UBCSAT output. We also want to keep track of the whole output file.
		run_cmd = "#{algorun_cmd} -runs #{numRuns} > #{algo_output_file}"
		run_cmd = "./" + run_cmd unless run_cmd =~/^\.\//
		puts "  Calling #{run_cmd}"
		system run_cmd

		require "ubcsat_reader.rb"
		readUbcsatAndWriteToDB(algo_output_file, algorun_config_id)
		File.delete(algo_output_file)
	elsif algo == "sls4mpe"
		#=== SLS4MPE -> Call special code to parse non-trivial UBCSAT output. We also want to keep track of the whole output file.
		run_cmd = "#{algorun_cmd} -x #{numRuns} > #{algo_output_file}"
		run_cmd = "./" + run_cmd unless run_cmd =~/^\.\//
		puts "  Calling #{run_cmd}"
		system run_cmd

		require "sls4mpe_reader.rb"
		readSLS4MPEAndWriteToDB(algo_output_file, algorun_config_id)
		File.delete(algo_output_file)
	elsif algo == "branin_function" 
		#=== Branin function script -> simple output, just parse here.
		run_cmd = "#{algorun_cmd} #{numRuns}"
		puts "Calling: #{run_cmd}"
		mysql_cmds = []
		File.popen(run_cmd){|file|
			while line = file.gets
				puts line
				if line =~/Observed value (#{float_regexp})./
					result = $1
					if line =~ /Solved./
						solved = "SAT"
					else
						solved = "TIMEOUT"
					end
		        		mysql_cmds << "insert into ALGORUN (ALGORUN_CONFIG_ID,MEASURED_RUNTIME,SOLVED, DATE) VALUES (#{algorun_config_id}, #{result}, '#{solved}', now());"
				end
			end
		}
	        raise "No branin_function run generated by command #{run_cmd}}" unless mysql_cmds.length > 0
		execute_cmds(mysql_cmds)
	else    
		raise "Algo #{algo} not implemented yet."
	end
end

=begin
###########################################################################
# Summarize the results of an instanceHash.
# An instanceHash is a Hash {instance_name=>{"desired_qual"=>optqual, "result"=>[res_1, res_2, ..., res_numLsRuns]}} where res_i = [time, steps, solqual_found]
###########################################################################
def summarizeResults(algo, instanceHash, instances_sorted)
	require "stats_ils.rb"

	total_cputime = 0
	all_med_times = []
	all_med_steps = []
	all_approx_quals = []
	all_sol_found = []
	for instance in instances_sorted
		results = instanceHash[instance]["result"]
		runtimes = []
		runlengths = []
		found_quals = []
		for result in results
			total_cputime += result[0].to_f
			runtimes << result[0].to_f
			runlengths << result[1].to_f
			found_quals << result[2].to_f
		end
		all_med_steps << runlengths.median
		all_med_times << runtimes.median
		all_approx_quals << instanceHash[instance]["desired_qual"].to_f / found_quals.avg
		#p [found_quals, instanceHash[instance]["desired_qual"].to_f, found_quals.avg/instanceHash[instance]["desired_qual"].to_f]
	end
	return [all_med_steps, all_med_times, all_approx_quals, total_cputime]
end

###########################################################################
# Returns the quality of a set of results for a particular algorithm.
###########################################################################
def objectiveFunction(algo, instanceHash)
	all_med_steps, all_med_times, all_approx_quals, total_cputime = summarizeResults(algo, instanceHash, instanceHash.keys)
	if algo == "saps"
		return [all_med_steps.median, total_cputime]
	elsif algo == "branin_function"
		return [all_med_times.median, total_cputime]
	elsif	algo == "sls4mpe"
		#p all_approx_quals
		return [1-all_approx_quals.avg, total_cputime] #[100*(1-all_approx_quals.avg), total_cputime]
	else
		raise "Need to implement objective function for algorithm #{algo}."
	end
end

###########################################################################
# For each run in the instance hash, return what objectiveFunction would be for only that run, in the order of instances in instance_array (instances can be repeated).
###########################################################################
def objectiveFunctions(algo, iHash, instance_array)
	finalresult = []
	instanceHash = iHash.deep_clone
	for instance in instance_array
		my_hash = {instance => {"result"=>[instanceHash[instance]["result"].shift], "desired_qual"=>instanceHash[instance]["desired_qual"]}}
		finalresult << objectiveFunction(my_hash)
	end
end

###########################################################################
# Outputs the objective function values for each of the instances -- mainly for CALIBRA.
# An instanceHash is a Hash {instance_name=>{"desired_qual"=>optqual, "result"=>[res_1, res_2, ..., res_numLsRuns]}} where res_i = [time, steps, solqual_found]
###########################################################################
def outputCalibraObjectives(objectives, algo)
	for obj in objectives
		if algo == "saps"
			puts "Calibra objective: #{1+all_med_steps[i]/1000000}"
		elsif algo == "branin_function"
			puts "Calibra objective: #{1+all_med_times[i].median}"
		elsif algo == "sls4mpe"
			puts "Calibra objective: #{1+1-all_approx_quals[i]}" #1 is perfect approx qual, 0 is worst
		else
			raise "Need to implement output function for algorithm #{algo}."
		end
	end
	puts "Total runtime: #{total_cputime}"
	obj, time = objectiveFunction(algo, instanceHash)
	puts "Total objective: #{obj}"
end
=end


###########################################################################
# Outputs the objective function values for each of the instances -- mainly for CALIBRA.
# An instanceHash is a Hash {instance_name=>{"desired_qual"=>optqual, "result"=>[res_1, res_2, ..., res_numLsRuns]}} where res_i = [time, steps, solqual_found]
###########################################################################
def outputObjectives(instances_sorted, objectives)
	raise "not a result for each instance: #{instances_sorted}, #{objectives}" unless instances_sorted.length == objectives.length

	for i in 0...objectives.length
		puts "#{instances_sorted[i]}: #{objectives[i][0]}\t\t\t\t[based on #{objectives[i][1]} runs]"
	end
end


def outputCalibraObjectivesFairEval(algo, within_obj, obj_function, instances_sorted, objectives)
	raise "not a result for each instance: #{instances_sorted}, #{objectives}" unless instances_sorted.length == objectives.length

	for i in 0...objectives.length
		if within_obj == "median_runtime"
			obj = objectives[i][0]+1
			obj = 1000 if obj > 1000
		elsif within_obj  == "median_runlength"
			if 1+objectives[i][0] > 1000000*1000
				obj = 1000
			else
				obj = 1 + objectives[i][0] / 1000000
			end
		elsif algo == "branin_function"
			obj = 1 + objectives[i][0] 
		elsif algo == "sls4mpe"
			obj = 1 + objectives[i][0] 
		else
			raise "Need to implement calibra output function for algorithm #{algo}."
		end
		
		puts "Calibra objective: #{obj}\n  [based on #{objectives[i][1]} runs for #{instances_sorted[i]}]"
	end
	puts "Total runtime: #{$total_cputime}"
end

###########################################################################
# Extract the appropriate value for the algorithm from the result [runtime, runlength, qual_found, des_qual]
###########################################################################
def singleInstanceObjective(algo, within_obj, results, qual)
	require "stats_ils.rb"
	return 100000000000000 if results.empty?
	max_runtime = 100000000000000
	max_runlength = 100000000000000
	eps = 0.0001
	runtimes = []
	runlengths = []
	found_quals = []
	for result in results
		runtime, runlength, found_qual = result.map{|x|x.to_f}
		if found_qual.to_f <= qual.to_f + eps
			runtimes << runtime
			runlengths << runlength
		else
			runtimes << max_runtime
			runlengths << max_runlength
		end
		found_quals << found_qual
	end

	return runlengths.median if within_obj == "median_runlength"
	return runtimes.median if within_obj == "median_runtime"
	if within_obj == "avg_approx"
		approx_qual = qual.to_f/found_quals.avg  # <= 1, want to max
		return 1 - approx_qual
	end
	raise "Need to define within_obj #{within_obj} in singleInstanceObjective in algo_specifics."
end

def getObjectivesForComputedInstances(algo, within_obj, instances, instances_sorted, state_as_string)
	#=== Results are stored directly in the instances array. Just extract them.
	runsWithResults = []
	resultsCanExist = true
	for entry in instances
		if entry["resultForState"].key?(state_as_string)
			raise "@instances array has no result for early instance but result for latter one: state #{state_as_string}, @instances #{@instances}" unless resultsCanExist
			runsWithResults << entry
		else
			resultsCanExist = false  #=== Just as a safety check.
		end
	end
	return getAllObjectives(algo, within_obj, runsWithResults, instances_sorted, state_as_string)
end

=begin
def getGlobalObjective(algo, instances, instances_sorted, state_as_string)
	totalNumInstances = 0
	instanceObjectives = getObjectivesForComputedInstances(algo, instances, instances_sorted, state_as_string)
	instanceObjectives.map{|x|totalNumInstances += x[1]}
	instanceObjectives.map!{|x|x[0]}
	
	#=== Combine objective functions.
	val = combinationOfObjectiveFunctions(algo, instanceObjectives)
	return [val, totalNumInstances]
end
=end

def getGlobalObjective(algo, within_obj, obj_function, instances, instances_sorted, state_as_string)
	require "stats_ils.rb"
	results = []
	for entry in instances
		if entry["resultForState"].key?(state_as_string)
			result = entry["resultForState"][state_as_string] 
			results << singleInstanceObjective(algo, within_obj, [result], entry["qual"])
		end
	end
#	return [results.avg, results.length]
	return [combinationOfObjectiveFunctions(algo, obj_function, results), results.length]
end

###########################################################################
# Extract the appropriate value for the algorithm from the result [runtime, runlength, qual_found, des_qual]
###########################################################################
def getAllObjectives(algo, within_obj, instances, instances_sorted, state_as_string)	
	instanceHash = makeStandardInstanceHash(instances, state_as_string)
	
	objectives = []
	for inst in instances_sorted
		if instanceHash.key?(inst)
			result = instanceHash[inst]["result"]
			qual = instanceHash[inst]["desired_qual"]
			objectives << [singleInstanceObjective(algo, within_obj, result, qual), result.length]
		else
			objectives << [nil, 0]
		end
	end
	return objectives
end

###########################################################################
# Summarize the results as one float.
###########################################################################
def combinationOfObjectiveFunctions(algo, obj_function, instanceObjectives)
	return 100000000000000 if instanceObjectives.empty?
	
	require "stats_ils.rb"
	return instanceObjectives.quantile(0.9) if obj_function == "q90"
	return instanceObjectives.median if obj_function == "median"
	return instanceObjectives.avg if obj_function == "mean"
	return instanceObjectives.avg if obj_function == "avg"
	
	raise "Need to implement combination of objective functions #{obj_function}."
end


=begin
def outputCalibraObjectives(algo, instances_sorted, objectives)
	raise "not a result for each instance: #{instances_sorted}, #{objectives}" unless instances_sorted.length == objectives.length

	for i in 0...objectives.length
		if algo == "saps"
			if 1+objectives[i][0] > 1000000*1000
				obj = 1000
			else
				obj = 1 + objectives[i][0] / 1000000
			end
		elsif algo == "branin_function"
			obj = 1 + objectives[i][0] 
		elsif algo == "sls4mpe"
			obj = 1 + objectives[i][0] 
		else
			raise "Need to implement calibra output function for algorithm #{algo}."
		end
		
		puts "Calibra objective: #{obj}\n  [based on #{objectives[i][1]} runs for #{instances_sorted[i]}]"
	end
	puts "Total runtime: #{$total_cputime}"
	objectives.map!{|x|x[0]}
	puts "Total objective: #{combinationOfObjectiveFunctions(algo, objectives)}"
end
=end